"""Test models for unit testing"""

import logging
from simu import Model
from simu.core.thermo.material import MaterialSpec
from simu.core.utilities import SymbolQuantity


class SimpleParameterTestModel(Model):
    """A simple model to test parameters"""

    def interface(self):
        with self.parameters as p:
            p.define("length", 10, "m")

    def define(self):
        pass


class ParameterTestModel(Model):
    """A simple model to test parameters"""

    def interface(self):
        with self.parameters as p:
            p.define("length", 10, "m")
            p.define("width", unit="m")

    def define(self):
        par = self.parameters
        area = par["length"] * par["width"]
        logging.debug(f"area = {area:~}")


class PropertyTestModel(Model):
    """A simple model to test properties"""

    def interface(self) -> None:
        self.parameters.define("length", 10, "m")
        self.properties.declare("area", "m^2")

    def define(self) -> None:
        self.properties["area"] = self.parameters["length"] ** 2


class HierarchyTestModel(Model):
    """A simple hierarchical model, where the child module calculates the
    square (surface) of a parameter, and the parent model calculates the
    volume as function of the calculated surface and a depth parameter."""

    def interface(self):
        self.parameters.define("depth", 5.0, "cm")
        self.properties.declare("volume", unit="m**3")

    def define(self):
        with self.hierarchy.add("square", PropertyTestModel) as child:
            pass
        volume = child.properties["area"] * self.parameters["depth"]
        self.properties["volume"] = volume


class HierarchyTestModel2(Model):
    """A simple hierarchical model, where the parent model calculates the
    length of a parameter, and the child model calculates the
    surface as function of the calculated length."""

    def interface(self):
        self.hierarchy.declare("square", PropertyTestModel)
        with self.parameters as p:
            p.define("radius", 5.0, "cm")
            p.define("depth", 10, "cm")
        self.properties.declare("volume", "m**3")

    def define(self):
        radius = self.parameters["radius"]
        length = 2 * radius
        # # later shortcut syntax proposal:
        # with self.child("square", SquareTestModel) as child:
        with self.hierarchy.add("square", PropertyTestModel) as child:
            child.parameters.provide(length=length)

        volume = self.parameters["depth"] * child.properties["area"]
        self.properties["volume"] = volume


class MaterialTestModel2(Model):
    def interface(self):
        spec = MaterialSpec(["H2O", "*"])
        self.materials.define_port("inlet", spec)

    def define(self):
        inlet = self.materials["inlet"]
        _ = self.materials.create_state("local", inlet.definition)


class ResidualTestModel(Model):
    """A simple model to test residuals"""
    def interface(self):
        self.parameters.define("length", 10, "m")
        self.parameters.define("area_spec", 50, "m**2")

    def define(self):
        res = self.parameters["length"] ** 2 - self.parameters["area_spec"]
        self.residuals.add("area", res, "m**2")


class ResidualTestModel2(Model):
    def interface(self):
        pass

    def define(self):
        res = SymbolQuantity("Hubert", "K")
        self.residuals.add("Hubert", res, "degC")


class BoundTestModel(Model):
    def interface(self):
        self.parameters.define("length", 10, "m")

    def define(self):
        length = self.parameters["length"]
        self.bounds.add("length", length)


from .basic import (
    H0S0ReferenceState, LinearHeatCapacity, StandardState, IdealMix,
    GibbsIdealGas, HelmholtzIdealGas, ConstantGibbsVolume, MolecularWeight,
    ChargeBalance)
from .cubic import (
    NonSymmetricMixingRule, LinearMixingRule, BostonMathiasAlphaFunction,
    CriticalParameters, RedlichKwongEOSLiquid, RedlichKwongEOSGas,
    RedlichKwongAFunction, RedlichKwongBFunction, RedlichKwongMFactor,
    VolumeShift)
from .special import Derivative

all_contributions = [H0S0ReferenceState, LinearHeatCapacity, StandardState,
                     IdealMix, GibbsIdealGas, HelmholtzIdealGas,
                     ConstantGibbsVolume, MolecularWeight, ChargeBalance,
                     NonSymmetricMixingRule, LinearMixingRule,
                     BostonMathiasAlphaFunction, CriticalParameters,
                     RedlichKwongEOSLiquid, RedlichKwongEOSGas,
                     RedlichKwongAFunction, RedlichKwongBFunction,
                     RedlichKwongMFactor, VolumeShift, Derivative]



"""This module contains the base classes to represent (process) models.

"""
from abc import ABC, abstractmethod
from typing import Self

from .parameter import ParameterHandler, ParameterProxy
from .hierarchy import HierarchyHandler, HierarchyProxy
from .property import PropertyHandler, PropertyProxy
from .material import MaterialHandler, MaterialProxy
from .bound import BoundHandler, BoundProxy
from simu.core.utilities.residual import ResidualHandler, ResidualProxy


class Model(ABC):
    """This is the base class for all process models to be implemented.

    By deriving from this class, handler objects are available as class
    attributes to deal with the particular aspects. Visit their documentation
    for details.

    The model implementation is then divided into two parts: (a) the interface,
    and (b) the model implementation itself. These parts are represented by the
    two methods :meth:`interface` and :meth:`define` respectively.

    A model is then either instantiated by a parent model, whereas it is
    represented in that parent model by a
    :class:`~simu.core.model.base.ModelProxy` object, or it can be defined as
    the top level model by calling :meth:`top`.
    """

    parameters: ParameterHandler
    """The handler object that takes care of parameter configuration"""

    properties: PropertyHandler
    """The handler object that takes care of property configuration"""

    hierarchy: HierarchyHandler
    """The handler object that takes care of defining sub models"""

    materials: MaterialHandler
    """The handler object that takes care of materials"""

    residuals: ResidualHandler = None
    """The handler object that takes care of residuals"""

    bounds: BoundHandler = None
    """The handler object that takes care of domain boundaries"""

    def __init__(self):
        """The constructor is parameterless but still needs to be called by the
         subclass constructors (if implemented) to initialise the object.
         Defining a constructor for the subclasses can be useful to pass custom
         data into the model.
         """
        self.__proxy = None
        self.parameters = ParameterHandler(self.cls_name)
        self.properties = PropertyHandler()
        self.materials = MaterialHandler()
        self.hierarchy = HierarchyHandler(self)
        self.interface()
        self.residuals = ResidualHandler()
        self.bounds = BoundHandler()

    @classmethod
    def top(cls, name: str = "model") -> "ModelProxy":
        """Define this model as top level model, hence instantiate,
        create proxy, and finalise it.

        This is the recommended and fastest way to declare a model as being the
        top level model. It will create a proxy object (via :meth:`proxy`) and
        finalise the configuration of it via
        :meth:`simu.core.model.base.ModelProxy.finalise`.

        Performing the finalisation in one go after creating the proxy object
        is only possible, if the model is suitable to be top model. That is,
        it must not have material ports or parameters to be connected.

        :param name: The name of the top level model
        :type name: str
        :return: The readily configured
          :class:`~simu.core.model.base.ModelProxy` object
        """
        return cls.proxy(name).finalise()

    @classmethod
    def proxy(cls, name: str = "model") -> "ModelProxy":
        """Instantiate and create proxy of this model.

        This is a helper method called by the
        :class:`~simu.core.model.hierarchy.HierarchyHandler`
        when defining a child model. After the proxy class is generated,
        the parent model can connect materials and provide parameters. When
        that is done, the proxy object can be finalised.

        :param name: The name of the top level model
        :type name: str
        :return: The :class:`~simu.core.model.base.ModelProxy` object, ready
          for configuration from parent context. It then still needs to be
          finalised.
        """
        return cls().create_proxy(name)

    def interface(self) -> None:
        """This virtual method is to define all model parameters, material
        ports, and properties provided by this model. This makes the interface
        of the model in the hierarchical context nearly self-documenting.
        A simple example implementation could be

        .. code-block::

            def interface(self):
                \"\"\"Here a nice documentation of the model interface\"\"\"
                self.parameters.define("length", 10.0, "m")
                self.properties.provide("area", unit="m**2")

        Above interface requires a parameter called ``length`` with a default
        value of 10 metres. It promises to calculate a property called ``area``
        which has a unit compatible to square metres.
        """

    @abstractmethod
    def define(self) -> None:
        """This abstract method is to define the model implementation,
        including the use of submodules, creation of internal materials, and
        calculation of residuals and model properties. Matching to the example
        described in the :meth:`interface` method, a simple implementation
        could be

        .. code-block::

            def define(self):
                \"\"\"Here documentation of the internal function of the model.
                This distinction can be used to include this doc-string only
                for detailed documentation sections.\"\"\"
                length = self.parameters["length"]
                self.properties["area"] = length * length

        Here we read out the previously defined parameter ``length`` and
        calculate the property ``area``.
        """

    # following the two methods finalise and create_proxy for explicit use,
    # if the context manager is not used.

    def create_proxy(self, name: str = "model") -> "ModelProxy":
        """Create a proxy object for configuration in hierarchy context. This
        is the instance variant of the :meth:`proxy` class method.
        """
        return ModelProxy(self, name)

    @classmethod
    @property
    def cls_name(cls) -> str:
        """The name of the derived class with module path for hopefully
        unique identification.

        This method is mainly used to uniquely name static parameters of
        a model.

        :return: The name of the class with the module path
        :rtype: str
        """
        module_name = cls.__module__
        cls_name = cls.__name__
        if module_name == "__main__":
            return f"{cls_name}"
        else:
            return f"{module_name}.{cls_name}"


class ModelProxy:
    """Proxy class for models, being main object to relate to when accessing
    a child model during definition of the parent model.

    As for the :class:`Model` class, this proxy version offers via its handlers
    functionality to deal with parameters, properties, hierarchy, materials,
    and residuals, but the angle of view is different:

    The ``Model`` class deals with the implementation of the model, while the
    ``ModelProxy`` class offers its access to connect to it as a client. This
    client is most likely a parent ``Model`` or, if it is a top level model,
    a :class:`~simu.NumericHandler` object.

    ``ModelProxy`` objects are created from within the :class:`Model` class,
    and do not need to be instantiated directly by `SiMu` client code.
    """

    parameters: ParameterProxy
    """The proxy of the parameter handler, to connect and update parameters"""

    properties: PropertyProxy
    """The proxy of the property handler, making properties available"""

    hierarchy: HierarchyProxy
    """The proxy of the hierarchy handler, to parametrise child models"""

    materials: MaterialProxy
    """The proxy of the material handler, to connect material ports"""

    residuals: ResidualProxy
    """A handler object that takes care of residuals. This is really just a
    non-mutable mapping of residuals, as the client code is not supposed to
    temper with the definition of the child model. The residuals are still
    browsable, as required for instance by the :class:`~simu.NumericHandler`
    object for obvious reasons."""

    bounds: BoundProxy
    """The non-mutable proxy of the bound handler, to be queried by the 
    :class:`~simu.NumericHandler`."""

    def __init__(self, model: Model, name: str):
        self.parameters = model.parameters.create_proxy()
        self.properties = model.properties.create_proxy()
        self.hierarchy = model.hierarchy.create_proxy()
        self.materials = model.materials.create_proxy()
        self.residuals = model.residuals.create_proxy()
        self.bounds = model.bounds.create_proxy()
        self.__model = model
        self.__set_name(name)

    def __set_name(self, name):
        """Set the name of the model for better error diagnostics"""
        self.parameters.set_name(name)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if exc_type is not None:
            return
        self.finalise()

    def finalise(self) -> Self:
        """After the parent model has connected parameters and materials,
        this method is called to define and finalise itself.
        The final part of this process is to validate correct configuration and
        to clean up data structures.
        """
        self.parameters.finalise()  # parameters are final now
        self.materials.finalise()  # all ports are connected
        self.__model.define()
        self.properties.finalise()  # properties can be queried now
        self.hierarchy.finalise()  # all declared sub-models are provided
        return self



from pytest import fixture, mark

from simu import (
    Model, NumericHandler, StringDictThermoSource, SpeciesDefinition,
    ThermoParameterStore, InitialState, MaterialDefinition, Material,
    MaterialSpec)

from simu.app.thermo.factories import ExampleThermoFactory


@fixture(scope="session")
def material_model_function(material_test_model3):
    """Make a function out of a model defining materials"""
    proxy = material_test_model3.top()
    numeric = NumericHandler(proxy)
    args = numeric.function.arg_structure
    results = numeric.function.result_structure
    return args, results


@fixture(scope="session")
def material_h2o_rk_liq() -> Material:
    rk_liq = "Boston-Mathias-Redlich-Kwong-Liquid"
    factory = ExampleThermoFactory()
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, rk_liq)
    store = ThermoParameterStore()
    initial_state = InitialState.from_std(1)
    material_def = MaterialDefinition(frame, initial_state, store)
    material = material_def.create_flow()
    return material


@fixture(scope="session")
def material_def_h2o_with_param() -> MaterialDefinition:
    factory = ExampleThermoFactory()
    species_db = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species_db, "Ideal-Solid")
    store = ThermoParameterStore()

    source = StringDictThermoSource({
        'H0S0ReferenceState': {
            's_0': {'H2O': '0 J / K / mol '},
            'dh_form': {'H2O': '0 J / mol '},
            'T_ref': '298.15 K', 'p_ref': '1 bar'},
        'LinearHeatCapacity': {'cp_a': {'H2O': '50.0 J / K / mol '},
                               'cp_b': {'H2O': '0 J / K ** 2 / mol '}},
        'ConstantGibbsVolume': {'v_n': {'H2O': '18 ml / mol '}}})
    store.add_source("default", source)
    initial_state = InitialState.from_std(1)
    return MaterialDefinition(frame, initial_state, store)


@fixture(scope="session")
def thermo_param():
    data = {
        'H0S0ReferenceState': {
            's_0': {'CH3-(CH2)2-CH3': '0 J/(mol*K)',
                    'CH3-CH2-CH3': '0 J/(mol*K)'},
            'dh_form': {'CH3-(CH2)2-CH3': '0 kJ/mol',
                        'CH3-CH2-CH3': '0 kJ/mol'},
            'T_ref': '25 degC',
            'p_ref': '1 atm'},
        'LinearHeatCapacity': {
            'cp_a': {'CH3-(CH2)2-CH3': '98 J/(mol*K)',
                     'CH3-CH2-CH3': '75 J/(mol*K)'},
            'cp_b': {'CH3-(CH2)2-CH3': '0 J/(mol*K*K)',
                     'CH3-CH2-CH3': '0 J/(mol*K*K)'}},
        'CriticalParameters': {
            'T_c': {'CH3-(CH2)2-CH3': '425 K', 'CH3-CH2-CH3': '370 K'},
            'p_c': {'CH3-(CH2)2-CH3': '38 bar', 'CH3-CH2-CH3': '42.5 bar'},
            'omega': {'CH3-CH2-CH3': 0.199, 'CH3-(CH2)2-CH3': 0.153}},
        'MixingRule_A': {'T_ref': '25 degC'},
        'VolumeShift': {
            'c_i': {'CH3-(CH2)2-CH3': '0 m ** 3 / mol',
                    'CH3-CH2-CH3': '0 m ** 3 / mol'}},
        'BostonMathiasAlphaFunction': {
            'eta': {'CH3-CH2-CH3': 0, 'CH3-(CH2)2-CH3': 0}}
    }
    return StringDictThermoSource(data)


@fixture(scope="session")
def material_test_model_4(material_def_h2o_with_param):
    """Need to make a fixture for this model, as it needs another fixture
    for the material definition"""
    class  MaterialTestModel4(Model):
        def interface(self):
            pass

        def define(self):
            self.materials.create_flow("flow1", material_def_h2o_with_param)
            self.materials.create_flow("flow2", material_def_h2o_with_param)

    return MaterialTestModel4


@fixture(scope="session")
def material_parent_test_model():
    material_definition = simple_material_definition_function(["H2O", "NO2"])

    class MaterialParentTestModel(Model):
        class Child(Model):
            def __init__(self, mat_def):
                super().__init__()
                self.mat_def = mat_def

            def interface(self):
                self.materials.define_port("port")

            def define(self):
                self.materials.create_flow("m_child", self.mat_def)

        def define(self):
            mat_def = material_definition
            flow = self.materials.create_flow("m_parent", mat_def)
            with self.hierarchy.add("child", self.Child, mat_def) as child:
                child.materials.connect("port", flow)

    return MaterialParentTestModel


@fixture()
def square_test_model():
    material_definition = simple_material_definition_function(
        ["CH3-CH2-CH3", "CH3-(CH2)2-CH3"])

    class SquareTestModel(Model):
        def __init__(self):
            super().__init__()
            self.no2sol = material_definition

        def interface(self):
            with self.parameters as p:
                p.define("T", 10, "degC")
                p.define("p", 10, "bar")
                p.define("N", 1, "mol/s")
                p.define("x_c3", 10, "%")

        def define(self):
            flow = self.materials.create_flow("local", self.no2sol)  # 4 DOF
            flow["N"] = flow["n"].sum()
            flow["x"] = flow["n"] / flow["N"]

            param, radd = self.parameters, self.residuals.add
            radd("N", flow["N"] - param["N"], "mol/s")
            res = flow["N"] * param["x_c3"] - flow["n"]["CH3-CH2-CH3"]
            radd("x", res, "mol/s")
            radd("T", flow["T"] - param["T"], "K")
            radd("p", flow["p"] - param["p"], "bar")

            d_n = 2 * flow["n"]["CH3-(CH2)2-CH3"] - flow["n"]["CH3-CH2-CH3"]
            self.bounds.add("delta n", d_n)

    return SquareTestModel


@fixture(scope="session")
def material_test_model3():
    material_definition = simple_material_definition_function(["H2O", "NO2"])


    class MaterialTestModel3(Model):
        def interface(self):
            pass

        def define(self):
            self.materials.create_flow("local", material_definition)

    return MaterialTestModel3


@fixture(scope="session")
def material_test_model():
    material_definition = simple_material_definition_function(["H2O", "NO2"])

    class MaterialTestModel(Model):
        def interface(self):
            spec = MaterialSpec(["H2O", "*"])
            self.materials.define_port("inlet", spec)

        def define(self):
            _ = self.materials["inlet"]
            _ = self.materials.create_flow("local", material_definition)

    return MaterialTestModel


@fixture(scope="session")
def simple_material_definition(request) -> MaterialDefinition:
    """Defines a material to use. Normally, this would be a singleton somewhere
    in the project."""
    return simple_material_definition_function(request.param)


def simple_material_definition_function(species) -> MaterialDefinition:
    """Defines a material to use. Normally, this would be a singleton somewhere
    in the project."""
    rk_liq = "Boston-Mathias-Redlich-Kwong-Liquid"
    factory = ExampleThermoFactory()
    species_db = {s: SpeciesDefinition(s) for s in species}
    frame = factory.create_frame(species_db, rk_liq)
    store = ThermoParameterStore()
    initial_state = InitialState.from_cbar(10.0, 10.0, [1.0] * len(species))
    return MaterialDefinition(frame, initial_state, store)

@fixture(scope="session")
def model_with_residual():
    def material_definition():
        ideal_liq = "Ideal-Liquid"
        factory = ExampleThermoFactory()
        species_db = {s: SpeciesDefinition(s)
                      for s in "H2O Na:1+ Cl:1-".split()}
        frame = factory.create_frame(species_db, ideal_liq)
        store = ThermoParameterStore()
        initial_state = InitialState.from_cbar(10.0, 1.0, [1.0, 0.1, 0.1])
        return MaterialDefinition(frame, initial_state, store)

    class ModelWithThermoResidual(Model):
        def interface(self):
            pass

        def define(self):
            definition = material_definition()
            self.materials.create_flow("liq", definition)

    return ModelWithThermoResidual


# -*- coding: utf-8 -*-
from logging import getLogger, NullHandler

from .core.utilities import (
    ParameterDictionary, base_magnitude, log, qsum, conditional, jacobian,
    qpow, qvertcat, sqrt, exp, log, log10, sin, cos, tan, arcsin, arccos,
    arctan, sinh, cosh, tanh, arcsinh, arccosh, arctanh, base_magnitude,
    Quantity, QuantityDict, MCounter, SymbolQuantity, QFunction, base_unit,
    flatten_dictionary, unflatten_dictionary, extract_units_dictionary,
    simplify_quantity, parse_quantities_in_struct, quantity_dict_to_strings
)

from .core.utilities.constants import (
    PI, R_GAS, V_LIGHT, H_PLANCK, GAMMA_G, N_A, ALPHA,
    SIGMA, F, EPS_0, MU_0, E_0, K_B, STD_GRAVITY
)

from .core.thermo import (
    InitialState, SpeciesDefinition, SpeciesDB, StateDefinition,
    ThermoContribution, ThermoFactory, ThermoFrame, ThermoParameterStore,
    AbstractThermoSource, NestedDictThermoSource, StringDictThermoSource,
    MaterialDefinition, Material, MaterialSpec
)

from .core.model import Model, AModel, NumericHandler

from .core.solver import SimulationSolver

# versioning
from ._version import VERSION as __version__

# logging
logger = getLogger(__name__)
logger.addHandler(NullHandler())



from pathlib import Path
from pprint import pprint
from yaml import safe_load

from simu import (
    ThermoFactory, InitialState, MaterialDefinition, ThermoParameterStore,
    StringDictThermoSource, SpeciesDB)
from simu.app.thermo import all_contributions, GibbsState

CURRENT_DIR = Path(__file__).parent

# load species database
with open(CURRENT_DIR  / "species_db.yml") as file:
    species = SpeciesDB(safe_load(file))

# load model structure database
with open(CURRENT_DIR  / "thermo_model_structures.yml") as file:
    model_structures = safe_load(file)

# load thermodynamic parameter database
with open(CURRENT_DIR  / "ideal_gas_param.yml") as file:
    parameter_source = StringDictThermoSource(safe_load(file))

factory = ThermoFactory()
factory.register_state_definition(GibbsState)
factory.register(*all_contributions)

frame = factory.create_frame(species.get_sub_db(["Methane"]),
                             model_structures["simple_ideal_gas"])

initial_state = InitialState.from_si(400, 2e5, [1.0])
store = ThermoParameterStore()
ch4_ideal = MaterialDefinition(frame, initial_state, store)
missing_symbols = store.get_missing_symbols()

store.add_source("my_source", parameter_source)



# -*- coding: utf-8 -*-

# internal modules
from .frame import ThermoFrame
from .factory import ThermoFactory
from .contribution import ThermoContribution
from .state import StateDefinition, InitialState
from .species import SpeciesDB, SpeciesDefinition

from .parameters import (
    ThermoParameterStore, AbstractThermoSource, NestedDictThermoSource,
    StringDictThermoSource
)

from .material import MaterialDefinition, Material, MaterialSpec


from time import time
from pathlib import Path
from numpy import logspace, zeros, log10, ravel
from numpy.random import random, seed
from scipy.sparse import random as srandom, csc_array
from scipy.sparse.linalg import spsolve
from pypardiso import spsolve as pspsolve
from casadi import DM, solve, Sparsity
from matplotlib import pyplot

N = 25000
NUM = 100
TIME_LIMIT = 10
FIG_PATH = Path(__file__).parents[3] / "doc" / "source" / "figures"

SCIPY, CASADI, CAS_SCI, PYPAR = range(4)

def create_objects(size, density):
    a_s = srandom(size, size, density, format="csc", dtype=float)
    a_s.setdiag(random(size) + size)
    b = random(size)
    sparsity = Sparsity(size, size, a_s.indptr, a_s.indices, True)
    a_c = DM(sparsity, a_s.data)
    return a_s, a_c, b

def conv_solve(a, b):
    # Convert from Casadi, solve in Scipy
    a = csc_array(a)
    spsolve(a, b)


def main():
    def measure(a, b, which):
        nonlocal functions
        func = functions[which]
        if func is None:
            return float("nan")
        start = time()
        func(a, b)
        result = time() - start
        if result > TIME_LIMIT:
            functions[which] = None
        return result

    functions = {
        SCIPY: spsolve,
        CASADI: solve,
        CAS_SCI: conv_solve,
        PYPAR: pspsolve}

    seed(0)  # reproducible random numbers
    sizes = logspace(1, log10(N), num=NUM)
    densities = [0.01, 0.02]
    times = zeros((NUM, 4, 2))
    for k, size in enumerate(sizes):
        size = int(size)
        active = [i for i, f in functions.items() if f is not None]
        if not active:
            times[k:, :, :] = float("nan")
            break
        for d, density in enumerate(densities):
            a_s, a_c, b = create_objects(size, density)
            times[k][SCIPY][d] = measure(a_s, b, SCIPY)
            times[k][CASADI][d] = measure(a_c, b, CASADI)
            times[k][CAS_SCI][d] = measure(a_c, b, CAS_SCI)
            times[k][PYPAR][d] = measure(a_s, b, PYPAR)
        print(k, size, ravel(times[k, :, 1]))


    pyplot.loglog(sizes, times[:, SCIPY, 0], "k-", label="Scipy, $\\varrho=0.01$")
    pyplot.loglog(sizes, times[:, CASADI, 0], "r-", label="Casadi")
    pyplot.loglog(sizes, times[:, CAS_SCI, 0], "g-", label="Casadi $\\to$ Scipy")
    pyplot.loglog(sizes, times[:, PYPAR, 0], "b-", label="pypardiso")
    pyplot.loglog(sizes, times[:, SCIPY, 1], "k--", label="$\\varrho=0.02$")
    pyplot.loglog(sizes, times[:, CASADI, 1], "r--")
    pyplot.loglog(sizes, times[:, CAS_SCI, 1], "g--")
    pyplot.loglog(sizes, times[:, PYPAR, 1], "b--")
    pyplot.grid()
    pyplot.legend(loc="best")
    pyplot.xlabel("System size")
    pyplot.ylabel("Computation time [s]")
    pyplot.savefig(FIG_PATH / "solver_comparison.png", bbox_inches="tight")
    pyplot.show()



if __name__ == '__main__':
    main()


from pytest import fixture
from simu import (SpeciesDefinition, SymbolQuantity, base_unit,
                  ParameterDictionary)
from simu.app import BostonMathiasAlphaFunction
from simu.core.utilities.types import Map


@fixture(scope="session")
def species_definitions_ab() -> Map[SpeciesDefinition]:
    """A simple example species definition map with 2 species, A and B"""
    return {"A": SpeciesDefinition("N2"),
            "B": SpeciesDefinition("O2")}


@fixture(scope="session")
def species_definitions_abc() -> Map[SpeciesDefinition]:
    """A simple example species definition map with 3 species, A, B and C"""
    return {"A": SpeciesDefinition("N2"),
            "B": SpeciesDefinition("O2"),
            "C": SpeciesDefinition("Ar")}


@fixture(scope="session")
def boston_mathias_alpha_function(species_definitions_ab):
    def sym(name: str, units: str) -> SymbolQuantity:
        return SymbolQuantity(name, base_unit(units))

    def vec(name: str, size: int, units: str) -> SymbolQuantity:
        return SymbolQuantity(name, base_unit(units), size)

    res = {
        "_m_factor": vec("m", 2, "dimless"),
        "_T_c": vec("T_c", 2, "K"),
        "T": sym("T", "K")
    }
    cont = BostonMathiasAlphaFunction(species_definitions_ab, {})
    cont.define(res)
    return res, cont.parameters





from abc import abstractmethod
from typing import Any
from inspect import currentframe, getargvalues

from .types import Map


class Configurable:
    """In particular the numerical solvers offer a number of options that can be
    provided by the constructor but also explicitly set. This class offers such
    functionality and supports validation of the individual options given."""

    __DEFAULT_VALIDATION = {"f": lambda x: True, "msg": ""}

    def __init__(self, exclude: list[str] = None):
        """
        The constructor uses the ``inspect`` module to obtain the arguments of
        the subclass constructor and collects them into the ``options``
        attribute after validation.

        Arguments given in ``exclude`` are not included here.
        """
        if exclude is None:
            exclude = []
        frame = currentframe().f_back
        args, _, _, values = getargvalues(frame)
        self._args = args[1:]
        self.options = {}
        for k in self._args:
            if k not in exclude:
                self.set_option(k, values[k])

    def set_option(self, name: str, value: Any):
        """
        Set the option with given ``name`` to the given ``value``. A
        ``KeyError`` is raised if there is no argument with given name, and a
        ``ValueError`` is raised if the value is not accepted.

        Only options defined as arguments in the constructor can be addressed
        with this method.
        """
        if name not in self._args:
            raise KeyError(f"{name} is not a possible option")
        validation = self._arg_validations.get(name, self.__DEFAULT_VALIDATION)
        if "replace_none" in validation and value is None:
            value = validation["replace_none"]
        if not validation["f"](value):
            raise ValueError(f"Option {name} is invalid: {validation["msg"]}")
        self.options[name] = value

    @property
    @abstractmethod
    def _arg_validations(self) -> Map[Map[Any]]:
        """The subclass must here provide a dictionary that maps each option
        name to a dictionary with the following keys:

        - ``f``: The validation function that is called with the value of the
          option and is to return ``True`` if the value is acceptable.
        - ``msg``: A specific error message to be included in the ``ValueError``
          thrown if ``f`` returned ``False``
        - ``replace_none`` (optional): A replacement value in case the value
          is ``None``. This can be used for mutable or complex default
          arguments.
        """
        ...

    @staticmethod
    def _validate_between(low: float , high: float):
        """This function can be used to define an argument validation
         between ``low`` and ``high``. It returns a dict with appropriate
         entries for ``f`` and ``msg`` as described for :meth:`_arg_validations`
         """
        return {
            "f": lambda x: low <= x <= high,
            "msg": f"must be between {low} and {high}"
        }



from yaml import safe_load
from pathlib import Path

from simu import (
    ThermoFactory, InitialState, MaterialDefinition, ThermoParameterStore,
    StringDictThermoSource, SpeciesDB)
from simu.app.thermo import all_contributions, all_states

CURRENT_DIR = Path(__file__).parent

class MaterialFactory:
    def __init__(self):
        # load species database
        with open(CURRENT_DIR / "species_db.yml") as file:
            self.species = SpeciesDB(safe_load(file))

        # load model structure database
        with open(CURRENT_DIR / "thermo_model_structures.yml") as file:
            self.model_structures = safe_load(file)

        # load thermodynamic parameter database
        with open(CURRENT_DIR / "ideal_gas_param.yml") as file:
            parameter_source = StringDictThermoSource(safe_load(file))

        self.store = ThermoParameterStore()
        self.store.add_source("my_source", parameter_source)

        factory = ThermoFactory()
        for state in all_states:
            factory.register_state_definition(state)
        factory.register(*all_contributions)
        self.factory = factory

    def create(self, model_name, species):
        frame = self.factory.create_frame(
            self.species.get_sub_db(species),
            self.model_structures[model_name])

        initial_state = InitialState.from_si(400, 2e5, [1.0] * len(species))
        return MaterialDefinition(frame, initial_state, self.store)


material_factory = MaterialFactory()

ch4_ideal_gas = material_factory.create("simple_ideal_gas", ["Methane"])


from simu import Model
from .material_factory import ch4_ideal_gas

class Source(Model):
    """A model of a methane source"""

    def interface(self):
        self.parameters.define("T", 25, "degC")
        self.parameters.define("p", 1, "bar")
        self.parameters.define("V", 10, "m^3/hr")

    def define(self):
        src = self.materials.create_flow("source", ch4_ideal_gas)
        self.residuals.add("T", self.parameters["T"] - src["T"], "K")
        self.residuals.add("p", self.parameters["p"] - src["p"], "bar")
        self.residuals.add("V", self.parameters["V"] - src["V"], "m^3/h")







# -*- coding: utf-8 -*-

# internal modules
from simu.core.utilities import assert_reproduction, SymbolQuantity, base_unit
from simu.app.thermo.contributions.special import Derivative


# auxiliary functions
def sym(name: str, units: str) -> SymbolQuantity:
    """Define a scalar symbolic quantity"""
    return SymbolQuantity(name, base_unit(units))


def vec(name: str, size: int, units: str) -> SymbolQuantity:
    """define a vector symbolic quantity"""
    return SymbolQuantity(name, base_unit(units), size)


def test_derivative():
    """Test definition of Derivative contribution"""
    tau = sym("tau", "dimless")
    C = vec("C", 3, "dimless")
    poly = C[0] + (C[1] + C[2] * tau) * tau
    res = {"T": tau, "poly": poly}
    opt = {"y": "poly", "x": "T"}
    deri = Derivative(["A", "B"], opt)
    deri.define(res)
    assert_reproduction(str(res["dpoly_dT"]))



from simu import AModel
from .material_factory import ch4_ideal_gas

class Source(AModel):
    """A model of a methane source"""

    def interface(self):
        self.pad("T", 25, "degC")
        self.pad("p", 1, "bar")
        self.pad("V", 10, "m^3/hr")

    def define(self):
        src = self.mcf("source", ch4_ideal_gas)
        self.ra("T", self.pa["T"] - src["T"], "K")
        self.ra("p", self.pa["p"] - src["p"], "bar")
        self.ra("V", self.pa["V"] - src["V"], "m^3/h")







from ..utilities import Quantity
from ..utilities.types import Map

BoundProxy = Map["Quantity"]


class BoundHandler(BoundProxy):
    """The boundary handler keeps process model properties that must be strictly
    positive in order to remain within the model's mathematical domain.

    Thermodynamic (material) properties, such as temperature, pressure and
    molar quantities do not need to be added, as their bounds are to be defined
    in the :class:`simu.ThermoContribution` object(s) that constraint(s) their
    domain.

    .. note::

        These bounds do not represent inequality constraints in an optimization
        problem, but true domain boundaries that need to be respected by any
        solver in order to keep a feasible vector of independent variables.

        It is also generally not a good idea to impose bounds that are based on
        the expected range for a solved model, as a solver might need to step
        through values that are not physically reasonable but yet within the
        model domain.

    Typical examples of bounds defined here are:

      - Temperature difference over a heat exchanger when using the logarithmic
        mean temperature. A temperature cross-over would cause negative
        arguments to the ``log`` function.
      - Process model parameters that might be subject to change, such as
        geometric dimensions of equipment.

    A **bad** example is using these bounds to keep a temperature value between
    a lower and an upper limit.
    """
    def __init__(self):
        self.__bounds = {}

    def add(self, name: str, bound: Quantity):
        """Add a quantity to the bound handler to signal the solver that its
        value must remain strictly positive.

        If the unit includes an offset, such as ``degC`` or ``barg``, this
        offset is eliminated:

        >>> from simu import SymbolQuantity
        >>> # in a real case, this better be a dependent property
        >>> b = SymbolQuantity("T", "degC")
        >>> handler = BoundHandler()
        >>> handler.add("T", b)
        >>> print(f"{handler["T"]}")
        T delta_degree_Celsius

        If the intention is really to define a bound at a given absolute value,
        this must explicitly be done as follows:

        >>> from simu import Quantity
        >>> handler.add("T2", b - Quantity(31.4159, "degC"))
        >>> print(f"{handler["T2"]}")
        (T-31.4159) delta_degree_Celsius
        """
        if name in self.__bounds:
            raise KeyError(f"Bound {name} already defined")

        # eliminate impact of offset in units like degC and barg
        self.__bounds[name] = bound - Quantity(0.0, bound.units)

    def __getitem__(self, key: str):
        return self.__bounds[key]

    def __len__(self):
        return len(self.__bounds)

    def __iter__(self):
        return iter(self.__bounds)

    def create_proxy(self) -> BoundProxy:
        """Create a proxy object"""
        return self



from yaml import safe_load

from ...core.thermo.parameters import StringDictThermoSource
from ..data import DATA_DIR


class ExampleThermoSource(StringDictThermoSource):
    def __init__(self):
        with open(DATA_DIR / "parameters.json") as file:
            data = safe_load(file)
        StringDictThermoSource.__init__(self, data)



# external modules
from casadi import vertsplit, vertcat

# internal modules
from ... import StateDefinition, Quantity
from ...core.utilities import base_unit, base_magnitude
from ...core.utilities.types import MutMap



class HelmholtzState(StateDefinition):
    """This definition interprets the state as being temperature, volume,
    and mole numbers. Accordingly, it defines:

    ======== ============================
    Property Description
    ======== ============================
    ``T``    Temperature
    ``V``    Volume
    ``n``    Mole vector
    ======== ============================
    """

    def prepare(self, result: MutMap[Quantity], flow: bool = False):
        state = result["_state"].magnitude
        result["T"], result["V"], *n_vec = vertsplit(state, 1)
        result["n"] = vertcat(*n_vec)
        s = "/s" if flow else ""
        for name, unit in [("T", "K"), ("V", f"m**3{s}"), ("n", f"mol{s}")]:
            result[name] = Quantity(result[name], base_unit(unit))

    def reverse(self, state):
        return [base_magnitude(state.temperature), None] + \
            list(base_magnitude(state.mol_vector))

    def declare_vector_keys(self, species):
        return {"n": list(species.keys())}

class GibbsState(StateDefinition):
    """This definition interprets the state as being temperature, pressure,
    and mole numbers. Accordingly, it defines:

    ======== ============================
    Property Description
    ======== ============================
    ``T``    Temperature
    ``p``    Pressure
    ``n``    Mole vector
    ======== ============================
    """

    def prepare(self, result: MutMap[Quantity], flow: bool = False):
        state = result["_state"].magnitude
        result["T"], result["p"], *n_vec = vertsplit(state, 1)
        result["n"] = vertcat(*n_vec)
        q_unit = "mol/s" if flow else "mol"
        for name, unit in [("T", "K"), ("p", "Pa"), ("n", q_unit)]:
            result[name] = Quantity(result[name], base_unit(unit))

    def reverse(self, state):
        return [base_magnitude(state.temperature),
                base_magnitude(state.pressure)] + \
                list(base_magnitude(state.mol_vector))

    def declare_vector_keys(self, species):
        return {"n": list(species.keys())}


all_states = [GibbsState, HelmholtzState]



from _pytest.python_api import raises
from pint import DimensionalityError, UndefinedUnitError

from simu.core.thermo import ThermoParameterStore
from simu.app.thermo.factories import ExampleThermoFactory
from simu.core.thermo.parameters import StringDictThermoSource
from simu.core.thermo.species import SpeciesDefinition
from simu.core.utilities import assert_reproduction


RK_LIQ = "Boston-Mathias-Redlich-Kwong-Liquid"


def test_create_frame():
    factory = ExampleThermoFactory()
    assert RK_LIQ in factory.structure_names
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, RK_LIQ)
    assert_reproduction(frame.parameter_structure)


def test_get_thermo_properties():
    factory = ExampleThermoFactory()
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, RK_LIQ)
    store = ThermoParameterStore()
    symbols = store.get_symbols(frame.parameter_structure)
    assert symbols["CriticalParameters"]["T_c"]["H2O"].units == "kelvin"


def test_get_thermo_properties_twice():
    factory = ExampleThermoFactory()
    species = {"H2O": SpeciesDefinition("H2O")}
    frame = factory.create_frame(species, RK_LIQ)
    store = ThermoParameterStore()
    symbols1 = store.get_symbols(frame.parameter_structure)
    symbols2 = store.get_symbols(frame.parameter_structure)
    qty1 = symbols1["CriticalParameters"]["T_c"]["H2O"]
    qty2 = symbols2["CriticalParameters"]["T_c"]["H2O"]
    assert qty1 is qty2


def test_get_thermo_properties_twice_wrong():
    store = ThermoParameterStore()
    struct_1 = {"A": {"B": "K"}}
    struct_2 = {"A": {"B": "m"}}
    store.get_symbols(struct_1)
    with raises(DimensionalityError):
        store.get_symbols(struct_2)


def test_thermo_source_get_item():
    struct = {"T": {"H2O": "100 K", "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    source = StringDictThermoSource(struct)
    t_h2o = source["T", "H2O"]
    assert f"{t_h2o:~}" == "100 K"


def test_thermo_source_try_nonleaf():
    struct = {"T": {"H2O": {"A": "1 K", "B": "2 K"}, "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    source = StringDictThermoSource(struct)
    with raises(KeyError):
        _ = source["T", "H2O"]


def test_thermo_source_try_non_quantity():
    struct = {"T": {"H2O": "Bite me!", "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    with raises(UndefinedUnitError):
        _ = StringDictThermoSource(struct)


def test_get_thermo_property_values_sources():
    store = ThermoParameterStore()
    struct = {"T": {"H2O": "K"}, "p": {"H2O": "bar"}}
    store.get_symbols(struct)

    struct = {"T": {"H2O": "100 K", "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    store.add_source("Dagbladet", StringDictThermoSource(struct))

    values = store.get_all_values()
    assert f"{values['T']['H2O']:~}" == "100 K"
    sources = store.get_sources()
    assert sources["T"]["H2O"] == "Dagbladet"


def test_get_thermo_missing():
    store = ThermoParameterStore()
    struct = {"T": {"CH4": "K", "H2O": "K"}}
    store.get_symbols(struct)

    struct = {"T": {"H2O": "100 K", "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    store.add_source("Dagbladet", StringDictThermoSource(struct))

    missing = store.get_missing_symbols()
    assert missing["T"]["CH4"].strip() == "K"


def test_get_missing_dimensionless():
    store = ThermoParameterStore()
    struct = {"T": {"CH4": "", "H2O": ""}}
    store.get_symbols(struct)
    missing = store.get_missing_symbols()
    assert len(missing["T"]) == 2


def test_get_thermo_property_values_two_sources():
    store = ThermoParameterStore()
    struct = {"T": {"H2O": "K"}, "p": {"H2O": "bar"}}
    store.get_symbols(struct)

    struct = {"T": {"H2O": "100 K", "NH3": "200 K"}}
    store.add_source("Dagbladet", StringDictThermoSource(struct))
    struct = {"p": {"H2O": "10 bar", "NH3": "20 atm"}}
    store.add_source("VG", StringDictThermoSource(struct))

    _ = store.get_all_values()
    sources = store.get_sources()
    assert sources["T"]["H2O"] == "Dagbladet"
    assert sources["p"]["H2O"] == "VG"


def test_get_same_thermo_property_values_two_sources():
    store = ThermoParameterStore()
    struct = {"T": {"H2O": "K"}, "p": {"H2O": "bar"}}
    store.get_symbols(struct)

    struct = {"T": {"H2O": "100 K", "NH3": "200 K"},
              "p": {"H2O": "10 bar", "NH3": "20 atm"}}
    store.add_source("Dagbladet", StringDictThermoSource(struct))
    struct = {"p": {"H2O": "10 bar", "NH3": "20 atm"}}
    store.add_source("VG", StringDictThermoSource(struct))

    _ = store.get_all_values()
    sources = store.get_sources()
    assert sources["T"]["H2O"] == "Dagbladet"
    assert sources["p"]["H2O"] == "VG"



# -*- coding: utf-8 -*-
"""Test module for governing thermo objects"""

# stdlib modules
from sys import argv
from pathlib import Path

# external modules
from pytest import main
from logging import DEBUG
from yaml import safe_load

# internal modules
from simu import ThermoFactory, InitialState, SpeciesDefinition
from simu.core.thermo import ThermoFactory
from simu.app.thermo.contributions import (
    H0S0ReferenceState, LinearHeatCapacity, StandardState, IdealMix,
    HelmholtzIdealGas)
from simu.app.thermo import HelmholtzState
from simu.core.utilities import (
    assert_reproduction, parse_quantities_in_struct, Quantity as Qty)


filename = Path(__file__).resolve().parent / "example_parameters.yml"
with open(filename, encoding="utf-8") as file:
    example_parameters = parse_quantities_in_struct(safe_load(file))


def test_create_thermo_factory():
    """just create a ThermoFactory"""
    ThermoFactory()


def test_register_contributions():
    """Create a ThermoFactory and register some contribtions"""
    create_frame_factory()


def test_create_frame(caplog):
    """create a ThermoFrame object"""
    fac = create_frame_factory()
    config = {
        "species": ["N2", "O2", "Ar", "CO2", "H2O"],
        "state": "HelmholtzState",
        "contributions": [
            "H0S0ReferenceState", "LinearHeatCapacity", "StandardState"
        ],
    }
    species = {f: SpeciesDefinition(f) for f in "N2 O2 Ar CO2 H2O".split()}
    with caplog.at_level(DEBUG, logger="simu"):
        fac.create_frame(species, config)
    msg = "\n".join([r.message for r in caplog.records])
    for contrib in config["contributions"]:
        assert contrib in msg


def test_parameter_structure():
    """Retrieve an (empty) parameter structure from created frame"""
    frame = create_simple_frame()
    assert_reproduction(dict(frame.parameter_structure))


def test_property_structure():
    """Retrieve the names of defined properties from created frame"""
    frame = create_simple_frame()
    assert_reproduction(frame.property_structure)


def test_call_frame_flow():
    """Call a created frame with numerical values"""
    result = call_frame(flow=True)[2]
    result = {k: v for k, v in result.items() if k in {"S", "mu"}}
    assert_reproduction(result)


def test_call_frame_state():
    """Call a created frame with numerical values"""
    result = call_frame(flow=False)[2]
    result = {k: v for k, v in result.items() if k in {"S", "mu"}}
    assert_reproduction(result)


def test_initial_state():
    """Test whether initialisation of a Helmholtz ideal gas contribution
    gives the correct volume"""
    frame = create_simple_frame()
    initial_state = InitialState(temperature=Qty("25 degC"),
                                 pressure=Qty("1 bar"),
                                 mol_vector=Qty([1, 1], "mol"))
    x = frame.initial_state(initial_state, example_parameters)
    assert_reproduction(x[1])


# *** helper functions

def call_frame(flow: bool = True):
    """Call a frame object with a state and return all with the result"""
    frame = create_simple_frame()
    state = Qty([398.15, 0.0448, 1, 1])  # = T, V, *n
    result = frame(state, example_parameters, flow=flow)
    return frame, state, result["props"]


def create_frame_factory():
    """Create a ThermoFactory and register standard state contributions"""
    fac = ThermoFactory()
    fac.register(H0S0ReferenceState, LinearHeatCapacity, StandardState,
                 IdealMix, HelmholtzIdealGas)
    fac.register_state_definition(HelmholtzState)
    return fac


def create_simple_frame():
    """Create a ThermoFrame based on just standard state contributions"""
    fac = create_frame_factory()
    config = {
        "species": ["N2", "O2"],
        "state": "HelmholtzState",
        "contributions": [
            "H0S0ReferenceState", "LinearHeatCapacity", "StandardState",
            "IdealMix", "HelmholtzIdealGas"
        ],
    }
    species = {"N2": SpeciesDefinition("N2"), "O2": SpeciesDefinition("O2")}
    return fac.create_frame(species, config)


if __name__ == "__main__":
    # only this file, very verbose and print stdout when started from here.
    main([__file__, "-v", "-v", "-s", "-rP"] + argv[1:])



# -*- coding: utf-8 -*-
"""
This module defines the :class:`~simu.ThermoFrame` and the
:class:`~simu.ThermoFactory` classes, of which the latter one is a factory for
the former.

:class:`~simu.ThermoFrame` objects represent thermodynamic models as state
function objects, calculating thermochemical properties as function of their
state and the model parameters.
"""
# stdlib modules
from typing import Type, Optional
from collections.abc import Mapping, Sequence
from logging import getLogger

# external modules
from casadi import SX

# internal modules
from .contribution import ThermoContribution
from .state import StateDefinition, InitialState
from .species import SpeciesDefinition
from ..utilities import (Quantity, ParameterDictionary, QFunction, qvertcat,
                         SymbolQuantity, extract_units_dictionary)
from ..utilities.types import NestedMap, Map, MutMap

ThermoContributionDict = Map[tuple[Type[ThermoContribution], Map]]
"""
A dictionary whose keys are the names of the contributions, and the values are
tuples of the belonging classes and the options belonging to the definition.
An example could be:

.. code-block::

    contribs = {
        "LinearHeatCapacity": (LinearHeatCapacity, {}),
        ...
        "MixingRule_A": (NonSymmetricMixingRule, {"target": "_ceos_a"}
    } 
    
This kind of structure is used to define the sequence of contributions in a 
:class:`~simu.ThermoFrame` object. Here, the class ``LinearHeatCapacity`` is
defined  straight forward with no options. For the mixing rule of the ``A`` 
contribution, called ``MixingRule_A``, it uses the ``NonSymmetricMixingRule``
class,  configured with ``ceos_a`` being the ``target``.

It is up to each individual :class:`~simu.ThermoContribution` implementation to
support and document their set of options.
"""

logger = getLogger(__name__)


class ThermoFrame:
    """This class represents the thermodynamic model, which defines a
    Function object :class:`QFunction` of the state vector and the parameter
    vector, and calculates a set of thermodynamic properties.

    The object should not be constructed via the class constructor by
    client code, but created by the :meth:`~simu.ThermoFactory.create_frame`
    method that handles the house-keeping of contribution classes.
    """

    def __init__(self, species: Map[SpeciesDefinition],
                 state_definition: StateDefinition,
                 contributions: ThermoContributionDict):
        """This constructor establishes a thermo frame function object
        with given species and contributions.
        """
        # need to instantiate the contributions
        contribs: Map[ThermoContribution] = {
            name: cls_(species, options)
            for name, (cls_, options) in contributions.items()
        }
        self.__species: Map[SpeciesDefinition] = species
        self.__vectors: MutMap[Sequence[str]] = {}

        def create_function(flow: bool = False):
            """Build up the result dictionary and define function"""
            params: MutMap[ParameterDictionary] = {}
            # define thermodynamic state (th, mc, [ch])
            state = SymbolQuantity("x", "dimless", len(species) + 2)
            # call the contributions; build up result dictionary
            result = {"_state": state}
            bounds = {}
            residuals = {}
            normed_residuals = {}
            state_definition.prepare(result, flow)
            self.__vectors.update(state_definition.declare_vector_keys(species))
            for name, c in contribs.items():
                logger.debug(f"Defining contribution '{name}'")
                c.reset()
                c.define(result)
                if c.vectors:
                    self.__vectors.update(c.vectors)
                if c.bounds:
                    bounds[name] = c.bounds
                if c.parameters:
                    params[name] = c.parameters
                if c.residuals:
                    residuals[name] = {
                        k: v.value for k, v in c.residuals.items()}
                    normed_residuals[name] = {
                        k: (v.value / v.tolerance).to("")
                        for k, v in c.residuals.items()}
            # create function
            args = {"state": state, "parameters": params}
            res = {"props": result, "bounds": bounds,
                   "residuals": residuals, "normed_residuals": normed_residuals}
            return QFunction(args, res, "thermo_frame"), params

        self.__function, parameters = create_function(flow=False)
        function, _ = create_function(flow=True)

        self.__res_units = {
            False: self.__function.res_units,
            True: function.res_units
        }

        self.__contributions: Map[ThermoContribution] = contribs
        self.__state_definition: StateDefinition = state_definition
        self.__default: Optional[InitialState] = None
        self.__param_struct: NestedMap[str] = \
            extract_units_dictionary(parameters)

    def __call__(self, state: SX | Sequence[float],
                 parameters: NestedMap[Quantity],
                 squeeze_results: bool = True, flow: bool = False):
        """Shortcut operator to call to the underlying function object.

        The function call can be a stand-alone evaluation of the thermodynamic
        model, given floating point quantities for the state and the
        parameters. Alternatively, called with `CasADi`_ ``SX`` based
        quantities to become part of a larger functional.

        :param state: A `CasADi`_ ``SX`` object or a sequence of floats,
          representing the thermodynamic state of the model. This is to be seen
          as a purely numerical object, as the physical interpretation, for
          instance as temperature, volume and mole flows, is first happening
          within the model.

        :param parameters: A nested dictionary with string keys and
          :class:`~simu.Quantity` leaves. Depending on the application, these
          quantities hold float or ``SX`` type magnitudes.

        :return: A list of property collections, representing the thermodynamic
          properties, in the structure as defined by :attr:`property_structure`.

        Depending on the ``flow`` parameter, extensive properties will be
        returned as flow quantities, such as ``W`` or ``mol/s`` or as stagnant
        state properties, such as ``J`` or ``mol``.
        """
        self.__function.res_units = self.__res_units[flow]
        return self.__function({
                "state": Quantity(state),
                "parameters": parameters
            }, squeeze_results)

    @property
    def species(self) -> Sequence[str]:
        """Returns a list of species names"""
        return list(self.__species.keys())

    @property
    def species_definitions(self) -> Map[SpeciesDefinition]:
        """Return the map of species definitions"""
        return self.__species

    @property
    def vector_keys(self) -> Map[Sequence[[str]]]:
        """Return the index keys for the registered vector properties.
        For a standard model, there should at least be entries for ``n`` and
        ``mu``."""
        return self.__vectors

    @property
    def property_structure(self) -> NestedMap[str]:
        """Returns a recursive structure properties, defining the calculated
        properties, bounds, and residuals as follows:

        .. code-block::

            {"props": ...,
             "bounds": ...,
             "residuals": ...,
             "normed_residuals": ...
             }

         The ``props`` sub-structure represents all calculated properties.
         The ``bounds`` sub-structure contains all bounds, represented as
           variables that need to be truly positive.
         The ``residuals`` structure contains the residuals in their given
           unit, while the ``normed_residuals`` structure contains the
           dimensionless ratios of residual values and tolerances.
         """
        return self.__function.result_structure

    @property
    def parameter_structure(self) -> NestedMap[str]:
        """This property is to aid the process of parametrizing a model.
        It returns the structure of all required model parameters. Initially,
        the returned object contains units of measurements that must be
        replaced with actual quantities (symbolic or not) before the
        function (:meth:`__call__`) or :meth:`initial_state` can be called .
        For the latter, float quantities have to be provided to the parameter
        object.
        """
        return self.__param_struct

    def create_symbol_state(self) -> SX:
        """Create a symbol state that can be used to call the object functional
        symbolically."""
        return SX.sym("x", len(self.__species) + 2)

    def initial_state(self, state: InitialState,
                      parameters: NestedMap[Quantity]) -> Sequence[float]:
        """Return a state estimate for given temperature, pressure and
        molar quantities - at given parameter set.

        This method queries all contributions top-down for an implementation
        of the initialise method. If not overwritten by any (and thus returning
        a state), the model is expected to be in Gibbs coordinates already,
        and (T, p, n) is the initial state.

        """
        st_def = self.__state_definition
        state_flat: Sequence[float] = st_def.reverse(state)
        if None not in state_flat:  # trivial case, Gibbs coordinates
            return state_flat

        # define the state, replacing non-explicitly given values with nan
        state_flat = [float("NaN") if x is None else x for x in state_flat]
        # calculate all properties ... accept NaNs, by calling own function
        properties = self(state_flat, parameters)["props"]

        # start from top and call contributions to try to initialise,
        #  based on results that might contain NaNs (if they depend on non-given state parts).
        for cont in reversed(self.__contributions.values()):
            result = cont.initial_state(state, properties)
            if result:
                result[0] = float(result[0])
                result[1] = float(result[1])
                return result
        msg = "No initialisation found for non-Gibbs surface"
        raise NotImplementedError(msg)



from numpy import squeeze
from numpy.testing import assert_allclose

from simu import NumericHandler, flatten_dictionary, Quantity, jacobian
from simu.examples.material_model import Source
from simu.core.utilities import assert_reproduction

from .models import *

def test_parameters():
    proxy = SimpleParameterTestModel.top()
    numeric = NumericHandler(proxy)
    args = numeric.function.arg_structure
    assert args[NumericHandler.MODEL_PARAMS]['length'] == 'm'


def test_properties():
    proxy = PropertyTestModel.top()
    numeric = NumericHandler(proxy)
    results = numeric.function.result_structure
    assert results[NumericHandler.MODEL_PROPS]['area'] == 'm ** 2'


def test_residuals():
    proxy = ResidualTestModel.top()
    numeric = NumericHandler(proxy)
    results = numeric.function.result_structure
    assert results[NumericHandler.RESIDUALS]['area'] == "m ** 2"


def test_material_collect_states(material_model_function):
    args = material_model_function[0]
    assert args["vectors"][NumericHandler.STATE_VEC] == ""


def test_material_collect_multiple_states(material_test_model_4):
    proxy = material_test_model_4.top()
    numeric = NumericHandler(proxy)
    state = numeric.arguments["vectors"][NumericHandler.STATE_VEC]
    assert len(state.magnitude.nz) == 6


def test_material_collect_props(material_model_function):
    results = material_model_function[1]
    assert_reproduction(results["thermo_props"]["local"])


def test_material_collect_thermo_param(material_model_function):
    args = material_model_function[0]
    assert_reproduction(args["thermo_params"]["default"])


def test_hierarchy_collect_numerics():
    numeric = NumericHandler(HierarchyTestModel2.top())
    results = numeric.function.result_structure
    assert "area" in results["model_props"]["square"]


def test_square_model(square_test_model):
    numeric = NumericHandler(square_test_model.top())
    ref = {"args": numeric.function.arg_structure,
           "res": numeric.function.result_structure}
    assert_reproduction(ref)


def test_square_model_args(thermo_param, square_test_model):
    model = square_test_model()
    material = model.no2sol
    numeric = NumericHandler(model.create_proxy().finalise())
    material.store.add_source("default", thermo_param)
    struct = numeric.function.arg_structure
    args = numeric.arguments
    check_same_keys(struct, args)


def test_square_model_call(thermo_param, square_test_model):
    model = square_test_model()
    material = model.no2sol
    numeric = NumericHandler(model.create_proxy().finalise())
    material.store.add_source("default", thermo_param)
    args = numeric.arguments
    res = flatten_dictionary(numeric.function(args))
    res = {k: f"{v:.6f~}" for k, v in res.items()}
    assert_reproduction(res)


def test_jacobian(thermo_param, square_test_model):
    model = square_test_model()
    material = model.no2sol
    numeric = NumericHandler(model.create_proxy().finalise())
    material.store.add_source("default", thermo_param)
    jac_id = numeric.register_jacobian(NumericHandler.RES_VEC,
                                       NumericHandler.STATE_VEC)
    args = numeric.arguments
    result = numeric.function(args)
    dr_dx = result[NumericHandler.JACOBIANS][jac_id].magnitude
    assert_reproduction(dr_dx.tolist())


def test_collect_hierarchy_material(material_parent_test_model):
    proxy = material_parent_test_model.top()
    for port_props in (True, False):
        numeric = NumericHandler(proxy, port_properties=port_props)
        ref = {"args": numeric.function.arg_structure,
               "res": numeric.function.result_structure}
        assert_reproduction(ref, suffix=f"{port_props}".lower())


def test_extract_parameters(thermo_param, square_test_model):
    model = square_test_model()
    store = model.no2sol.store
    numeric = NumericHandler(model.create_proxy().finalise())
    store.add_source("default", thermo_param)
    params = {'model_params': {'N': 'mol / s', 'T': '°C',
                               'p': 'bar', 'x_c3': '%'}}
    numeric.extract_parameters("param", params)
    names = numeric.vector_arg_names("param")
    jac_id = numeric.register_jacobian(NumericHandler.RES_VEC, "param")
    args = numeric.arguments
    result = numeric.function(args)
    dr_dp = result[NumericHandler.JACOBIANS][jac_id].magnitude
    ref = {"names": names, "J": dr_dp.tolist()}
    assert_reproduction(ref)


def test_collect_properties(thermo_param, square_test_model):
    model = square_test_model()
    store = model.no2sol.store
    numeric = NumericHandler(model.create_proxy().finalise())
    store.add_source("default", thermo_param)
    props = {'thermo_props': {'local': {'mu': {
        'CH3-(CH2)2-CH3': 'kJ/mol', 'CH3-CH2-CH3': 'kJ/mol'}}}}
    numeric.collect_properties("mu", props)
    names = numeric.vector_res_names("mu")
    jac_id = numeric.register_jacobian("mu", NumericHandler.STATE_VEC)
    args = numeric.arguments
    result = numeric.function(args)
    dmu_dx = result[NumericHandler.JACOBIANS][jac_id].magnitude
    ref = {"names": names, "J": dmu_dx.tolist()}
    assert_reproduction(ref)


def test_export_state(square_test_model):
    numeric = NumericHandler(square_test_model.top())
    state = numeric.export_state()
    assert_reproduction(state)


def test_import_state(square_test_model):
    model = square_test_model()
    numeric = NumericHandler(model.create_proxy().finalise())
    state = {
        'thermo': {'local': {
            'T': '100 °C', 'p': '5 bar',
            'n': {'CH3-CH2-CH3': '2 mol', 'CH3-(CH2)2-CH3': '1 mol'}}},
        'non-canonical': {}}
    species = list(state["thermo"]["local"]["n"].keys())
    numeric.import_state(state)
    state = model.materials["local"].initial_state.to_dict(species)
    assert_reproduction(state)


def test_retain_initial_values(thermo_param, square_test_model):
    model = square_test_model()
    numeric = NumericHandler(model.create_proxy().finalise())
    material = model.materials["local"]
    material.definition.store.add_source("default", thermo_param)
    params = numeric.arguments["thermo_params"]
    state = [283.15, 2 * 0.000196732, 2, 2]
    numeric.retain_state(state, params)
    pressure = material.initial_state.pressure
    assert Quantity(0.999, "MPa") < pressure < Quantity(1.001, "MPa")


def test_retain_and_args(thermo_param, square_test_model):
    model = square_test_model()
    numeric = NumericHandler(model.create_proxy().finalise())
    material = model.materials["local"]
    material.definition.store.add_source("default", thermo_param)
    params = numeric.arguments["thermo_params"]
    state = [283.15, 2 * 0.000196732, 2, 2]
    numeric.retain_state(state, params)
    new_state  = squeeze(numeric.arguments["vectors"]["states"].magnitude)
    assert_allclose(new_state, state)

def test_thermo_residual(model_with_residual):
    numeric = NumericHandler(model_with_residual.top())
    rs = numeric.function.result_structure
    assert rs[numeric.RESIDUALS]["liq"]["ChargeBalance"]["balance"] == "A"


def test_query_bounds():
    numeric = NumericHandler(Source.top())
    res = numeric.vector_res_names(numeric.BOUND_VEC)
    assert_reproduction(res)

def test_model_bounds():
    numeric = NumericHandler(BoundTestModel.top())
    res = numeric.vector_res_names(numeric.BOUND_VEC)
    assert_reproduction(res)


def test_bound_sensitivity():
    numeric = NumericHandler(Source.top())
    args = numeric.arguments
    names = numeric.vector_arg_names(numeric.STATE_VEC)
    state = SymbolQuantity("x", "", names)
    args[numeric.VECTORS][numeric.STATE_VEC] = state
    res = numeric.function(args, squeeze_results=False)
    res = res[numeric.VECTORS][numeric.BOUND_VEC]
    jac = jacobian(res, state).magnitude
    assert_reproduction(str(jac))


def test_vector_bound(square_test_model):
    numeric = NumericHandler(square_test_model.top())
    res = numeric.vector_res_names(numeric.BOUND_VEC)
    res = [r for r in res if r.startswith("local/IdealMix/")]
    ref = ["local/IdealMix/n/CH3-(CH2)2-CH3", "local/IdealMix/n/CH3-CH2-CH3"]
    assert res == ref


def check_same_keys(dic1, dic2):
    """Check whether the two nested dictionaries have the same keys"""
    def is_it(d):
        try:
            d.items()
        except AttributeError:
            return False
        return True

    assert is_it(dic1) == is_it(dic2)
    if not is_it(dic1):
        return
    assert set(dic1.keys()) == set(dic2.keys())
    for key, child in dic1.items():
        check_same_keys(child, dic2[key])




from simu import (
    ThermoFactory, SpeciesDefinition, parse_quantities_in_struct,
    InitialState
)
from simu.app.thermo import all_contributions, GibbsState

factory = ThermoFactory()
factory.register_state_definition(GibbsState)
factory.register(*all_contributions)

config = {
    "state": "GibbsState",
    "contributions": [
        "H0S0ReferenceState",
        "LinearHeatCapacity",
        "IdealMix",
        "GibbsIdealGas"
    ],
}
species = {"Methane": SpeciesDefinition("CH4")}
frame = factory.create_frame(species, config)

parameters = parse_quantities_in_struct({
    'H0S0ReferenceState': {
        'T_ref': '25 degC',
        'dh_form': {'Methane': '-74.87 kJ/mol'},
        'p_ref': '1 bar',
        's_0': {'Methane': '188.66 J/K/mol'}},
    'LinearHeatCapacity': {
        'cp_a': {'Methane': '35.69 J/K/mol'},
        'cp_b': {'Methane': '50 mJ/K**2/mol'}}
})

result = frame([400, 1e5, 1.0], parameters)

tpn = InitialState.from_si(400, 2e5, [1.0])
state = frame.initial_state(tpn, parameters)
result2 = frame(state, parameters)



"""Module containing classes to describe materials (thermodynamic phases) in
the modelling context."""

# stdlib modules
from typing import Optional
from collections.abc import Iterator, Collection

# internal modules
from ..thermo.material import MaterialSpec, Material, MaterialDefinition
from ..utilities.types import Map, MutMap
from ..utilities.errors import DataFlowError


class MaterialHandler(Map[Material]):
    """The material handler maintains the thermodynamic states represented as
    flows and states. When a model is created, the ``interface`` method can be
    used to define material ports. In the ``with`` context, invoked by the
    parent model, defined ports are to be connected to ``Material`` instances
    in the parent context.
    Finally, the model can create further (local) material instances.
    """
    def __init__(self):
        self.__materials: MutMap[Material] = {}
        self.__ports: MutMap[MaterialSpec] = {}

    def define_port(self, name: str, spec: Optional[MaterialSpec] = None):
        """Define a material port of the given name and specification.
        The name must be unique in this context. If no ``spec`` is given,
        any material is accepted.
        """
        if name in self.__ports:
            raise KeyError(f"Material port '{name}' already defined")
        self.__ports[name] = MaterialSpec() if spec is None else spec

    def __getitem__(self, name: str) -> Material:
        """Re-obtain the material specification, avoiding the need to keep a
        holding variable in the client scope code."""
        return self.__materials[name]

    def __len__(self) -> int:
        return len(self.__materials)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__materials)

    def create_flow(self, name: str,
                    definition: MaterialDefinition) -> Material:
        """Create a Material as a flow in the local context."""
        return self.__register(name, definition.create_flow())

    def create_state(self, name: str,
                     definition: MaterialDefinition) -> Material:
        """Create a Material as a state in the local context."""
        return self.__register(name, definition.create_state())

    @property
    def ports(self) -> Map[MaterialSpec]:
        """All defined ports of the model"""
        return self.__ports

    def create_proxy(self) -> "MaterialProxy":
        """Create a proxy object for configuration in material context"""
        return MaterialProxy(self)

    def finalise(self, connections: Map[Material]):
        """Internal method called to provide connected ports."""
        self.__materials |= connections

    def __register(self, name: str, material: Material) -> Material:
        """Create a Material in the local context.
        """
        if name in self.__ports:
            raise KeyError(f"{name} is already defined as a port")
        if name in self.__materials:
            raise KeyError(f"{name} is already defined as a material")
        self.__materials[name] = material
        return material


class MaterialProxy(Map[MaterialSpec]):
    handler: MaterialHandler

    def __init__(self, handler: MaterialHandler):
        self.handler = handler
        self.__ports: MutMap[MaterialSpec] = dict(handler.ports)
        self.__connected: MutMap[Material] = {}

    def connect(self, name: str, material: Material):
        """Connect the material in local context to the port with given name
        of the child model"""
        if name not in self:
            raise KeyError(f"Port of name {name} is not defined")
        try:
            spec = self.__ports.pop(name)
        except KeyError:
            raise KeyError(f"Port of name {name} is already connected")
        if not spec.is_compatible(material):
            raise ValueError(f"Provided material on port {name} is "
                             "incompatible to the provided material object")
        self.__connected[name] = material

    def free_ports(self) -> Collection[str]:
        """Return collection of all ports that are yet free"""
        return self.__ports.keys()

    def __getitem__(self, name: str) -> MaterialSpec:
        """Re-obtain the material object, avoiding the need to keep a
        holding variable in the client scope code."""
        return self.handler.ports[name]

    def __len__(self) -> int:
        return len(self.handler.ports)

    def __iter__(self) -> Iterator[str]:
        return iter(self.handler.ports)

    def finalise(self):
        """check that all ports are connected"""
        if self.__ports:
            missing = ", ".join(self.__ports.keys())
            msg = f"The following ports are not connected: {missing}"
            raise DataFlowError(msg)
        self.handler.finalise(self.__connected)



from pytest import raises as pt_raises
from casadi import SX

from simu.core.utilities import (Quantity, SymbolQuantity, assert_reproduction,
                                 jacobian, qsum, log, exp, qsqrt, qpow, conditional,
                                 base_unit, QFunction, flatten_dictionary,
                                 unflatten_dictionary, extract_units_dictionary)


def test_symbol_quantity():
    """Test propagation of symbolic quantities"""
    x_1 = SymbolQuantity("x1", "m", ["A", "B"])
    x_2 = SymbolQuantity("x2", "kg", 2)
    a = SymbolQuantity("a", "1/s")
    y_1 = a * x_1
    y_2 = a * x_2
    assert_reproduction([y_1, y_2])


def test_quantity():
    """Test instantiation of quantities"""
    x = [
        Quantity(1, "cm"),
        Quantity("1 cm"),
        Quantity(1, "J/mol").to_base_units()
    ]
    x.append(Quantity(x[0]))
    assert_reproduction(x)


def test_jacobian():
    """Test Jacobian function for quantities"""
    x = SymbolQuantity("x1", "m", ["A", "B"])
    a = SymbolQuantity("a", "1/s")
    y = a * x
    z = jacobian(y, x)
    assert_reproduction(z)


def test_sum1():
    """Test sum function for quantities"""
    x = SymbolQuantity("x1", "m", "ABCDEFU")
    y = qsum(x)
    assert_reproduction(y)


def test_log():
    """Test log function for quantities"""
    x1 = SymbolQuantity("x1", "m", "AB")
    x2 = SymbolQuantity("x2", "cm", "AB")

    with pt_raises(TypeError):
        log(x1)

    z = log(x1 / x2)
    assert_reproduction(z)


def test_exp():
    """Test exp function for quantities"""
    x1 = SymbolQuantity("x1", "m", "AB")
    x2 = SymbolQuantity("x2", "cm", "AB")

    with pt_raises(TypeError):
        exp(x1)

    z = exp(x1 / x2)
    assert_reproduction(z)


def test_sqrt():
    """Test sqrt function for quantities"""
    x = SymbolQuantity("x", "m^2", "AB")
    z = qsqrt(x)
    assert_reproduction(z)


def test_pow():
    """Test pow function for quantities"""
    x1 = SymbolQuantity("x1", "m", "AB")
    x2 = SymbolQuantity("x2", "cm", "AB")

    with pt_raises(TypeError):
        qpow(x1, x2)

    z = qpow(x1 / x2, x2 / x1)
    assert_reproduction(z)


def test_conditional():
    """Test conditional function for branching"""
    x1 = SymbolQuantity("x1", "m", "AB")
    x2 = SymbolQuantity("x2", "cm", "AB")

    # try invalid condition
    cond: SX = x1 > x2  # this is not a boolean, but an SX node.
    z = conditional(cond, x1, x2)
    assert_reproduction(z)


def test_base_unit():
    """Test converting units into base-units"""
    candidates = [
        "cm", "hour", "kmol", "t/hr", "barg", "W/mol", "m**2/s", "C", "V", "pi"
    ]
    res = {c: base_unit(c) for c in candidates}
    assert_reproduction(res)


def test_qfunction():
    """test QFunction with flat parameters and results"""
    x = SymbolQuantity("x1", "m", ["A", "B"])
    a = SymbolQuantity("a", "1/s")
    y = a * x
    f = QFunction({"x": x, "a": a}, {"y": y})

    # and now with quantities
    x = Quantity([1, 2], "cm")
    a = Quantity("0.1 kHz")
    y = f({"x": x, "a": a})["y"]

    assert_reproduction(y)


def create_flat():
    orig = {"C": {"A": 1, "B": 2}, "A": 3}
    flat = flatten_dictionary(orig)
    return orig, flat


def test_simple_flatten():
    """Check that a simple dict is flattened as expected"""
    orig, flat = create_flat()
    assert_reproduction(flat)


def test_unflatten():
    """Check that original dict is reproduced by un-flattening"""
    orig, flat = create_flat()
    rep = unflatten_dictionary(flat)
    assert rep == orig


def test_flatten_escaped():
    """Check that (un-)flattening works with escaped separators"""
    orig = {"C": {"A/D": 1, "B": 2}, "A": 3}
    flat = flatten_dictionary(orig)
    assert_reproduction(flat)
    rep = unflatten_dictionary(flat)
    assert rep == orig


def test_q_function_nested():
    """test QFunction with nested parameters and results"""
    x = SymbolQuantity("x1", "m", ["A", "B"])
    a = SymbolQuantity("a", "1/s")
    y = a * x
    f = QFunction({"x": x, "b": {"a": a}}, {"z": {"y": y}})

    # and now with quantities
    x = Quantity([1, 2], "cm")
    a = Quantity("0.1 kHz")
    y = f({"x": x, "b": {"a": a}})["z"]["y"]
    assert_reproduction(y)


def test_extract_units_dictionary():
    """Test extraction of units from a nested dictionary"""
    x = Quantity([1, 2], "cm")
    a = Quantity("0.1 kHz")
    struct = {"x": x, "b": {"a": a}}
    units = extract_units_dictionary(struct)
    assert_reproduction(units)



"""
This module contains general helper functions that are useful on several
levels, while relying to maximal degree on standard python structures.
"""
from re import escape, split
from typing import TypeVar, Callable
from collections import Counter

# internal modules
from .types import NestedMap, MutMap, Map, NestedMutMap

_V = TypeVar("_V")
_R = TypeVar("_R")
FLATTEN_SEPARATOR = "/"  # separator when (un-)flattening dictionaries


class MCounter(Counter):
    """This is a slight extention of the ``Collections.Counter`` class
    to also allow multiplication with integers:

        >>> a = MCounter({"a": 1})
        >>> b = MCounter({"b": 1})
        >>> a + 2 * b
        MCounter({'b': 2, 'a': 1})
    """

    def __mul__(self, other):
        if not isinstance(other, int):
            raise TypeError("Non-int factor")
        return MCounter({k: other * v for k, v in self.items()})

    def __rmul__(self, other):
        return self * other  # call __mul__

    def __add__(self, other):
        return MCounter(super().__add__(other))

    def __pos__(self):
        return self


def flatten_dictionary(structure: NestedMap[_V], prefix: str = "") -> Map[_V]:
    r"""Convert the given structure into a flat list of key value pairs,
    where the keys are ``SEPARATOR``-separated concatonations of the paths,
    and values are the values of the leafs. Non-string keys are converted
    to strings. Occurances of ``SEPARATOR`` are escaped by ``\``.

    >>> d: NestedMap[int] = {"a": {"b": 1, "c": 2}, "d": {"e/f": 3}}
    >>> flatten_dictionary(d)
    {'a/b': 1, 'a/c': 2, 'd/e\\/f': 3}
    """
    try:
        items = structure.items()  # is this dictionary enough for us?
    except AttributeError:  # doesn't seem so, this is just a value
        return {prefix: structure}  # type: ignore

    result: MutMap[_V] = {}
    # must sort to create the same sequence every time
    # (dictionary might have content permuted)
    for key, value in sorted(items):
        key = str(key).replace(FLATTEN_SEPARATOR, rf"\{FLATTEN_SEPARATOR}")  # esc. separator
        key = f"{prefix}{FLATTEN_SEPARATOR}{key}" if prefix else key
        result.update(flatten_dictionary(value, key))
    return result


def unflatten_dictionary(flat_structure: Map[_V]) -> NestedMap[_V]:
    r"""This is the reverse of :func:`flatten_dictionary`, inflating the
    given one-depth dictionary into a nested structure.

    >>> d = {"a/b": 1, "a/c": 2, r"d/e\/f": 3}
    >>> unflatten_dictionary(d)
    {'a': {'b': 1, 'c': 2}, 'd': {'e/f': 3}}
    """
    result: NestedMutMap[_V] = {}

    def insert(struct, sub_keys, sub_value):
        """insert one element into the nested structure"""
        first = sub_keys.pop(0)
        if sub_keys:
            if first not in struct:
                struct[first] = {}
            insert(struct[first], sub_keys, sub_value)
        else:
            struct[first] = sub_value

    # split by non-escaped separators and unescape escaped separators
    regex = rf'(?<!\\){escape(FLATTEN_SEPARATOR)}'
    for key, value in flat_structure.items():
        keys = [
            k.replace(rf"\{FLATTEN_SEPARATOR}", FLATTEN_SEPARATOR)
            for k in split(regex, key)
        ]
        insert(result, keys, value)
    return result


def nested_map(structure: NestedMap[_V],
               function: Callable[[_V], _R]) -> NestedMap[_R]:
    """Apply a unary function to each leaf values of the given nested
    dictionary, and return the same structure with the function's values as
    leafs.
    >>> a = {'a': {'b': 1, 'c': 2}, 'd': {'e/f': 3}}
    >>> nested_map(a, lambda x: 2 * x)
    {'a': {'b': 2, 'c': 4}, 'd': {'e/f': 6}}
    """
    try:
        items = structure.items()
    except AttributeError:
        return function(structure)
    return {k: nested_map(v, function) for k, v in items}


"""This module defines exception types"""

from pint.errors import DimensionalityError, UndefinedUnitError


class DataFlowError(RuntimeError):
    """An error used when the data flow of objects is not configured correctly.
    """

class IterativeProcessFailed(RuntimeError):
    """Base-class for exceptions in case an iterative process failed."""

class IterativeProcessInterrupted(IterativeProcessFailed):
    """Exception raised when an iterative process, such as a numerical solver
     has been interrupted by user intervention, normally callback functions."""

class NonSquareSystem(ValueError):
    """Exception raised if a system that is to be square is not."""
    def __init__(self, variables: int, equations: int,
                 name: str = "system matrix"):
        excess = "equations" if equations > variables else "variables"
        delta = abs(variables - equations)
        msg = f"Non-square {name}: {variables} variables vs. " \
              f"{equations} equations; {delta} too many {excess}."
        super().__init__(msg)



from sys import stdout
from typing import Tuple

from .types import Map


class ProgressTableOutput:
    """Based on defined column headings and printing format strings for the
    table elements, print the provided data successively to the given stream.

    The header is not printed before the first row of data is provided. The
    space taken by the row elements then determines the individual column
    widths. Example:

    >>> from simu import  Quantity
    >>> from io import StringIO  # this is only required to make doctest happy.
    >>> stream = StringIO()  # else, we could just use the default stream here.

    >>> cols = {"m": ("Magnitude", "{:9.2g}"), "u": ("Unit", "{:>15s}")}
    >>> table = ProgressTableOutput(cols, stream=stream)
    >>> table.row(Quantity(20, "m/s"))
    >>> table.row(Quantity(10, "degC"))
    >>> print(stream.getvalue())
    Row Magnitude           Unit
    --- --------- --------------
      1        20 meter / second
      2        10 degree_Celsius

    """
    def __init__(self, columns: Map[Tuple[str, str]], row_dig : int|None = 3,
                 row_head = "Row", stream=stdout):
        """
        The constructor configures the table based on the following parameters:

        :param columns:  The keys of the dictionary are the names of the
          attributes to tabulate. THe value tuple represents the column header
          (first element) and the formatting string (second element).
        :param row_dig: The number of digits to be reserved for the row number.
          If ``None``, do not print the row number. Otherwise, the row will be
          the first column of the table.
        :param row_head: The heading for the row column.
        :param stream:  The used output stream. If ``None``, no output will be
          generated.
        """
        self.__cols = columns
        self._first = True
        self._row_fmt = None if row_dig is None else f"{{:{row_dig}d}}"
        self._row_head = row_head
        self._write = (lambda t: None) if (stream is None) else stream.write
        self._row = 1

    def row(self, data: object, row: int = None):
        """Print a data row, whereas the attributes are extracted from the given
        ``object``. If this is the first row of the table, the headers will be
        printed with it, and the column width thereby determined.

        :param data: The object, which must have the attributes as defined as
          keys in the ``columns`` mapping given to the constructor.
        :param row:  The row number to be printed. If omitted, the row number
          is incremented from the previous row, starting with one.
        """
        write = self._write
        row = self._row if row is None else row

        elem = [c[1].format(getattr(data, k)) for k, c in self.__cols.items()]
        row_txt = "" if self._row_fmt is None else self._row_fmt.format(row)

        if self._first:
            headings = [f"{c[0]:>{len(e)}s}"
                        for e, c in zip(elem, self.__cols.values())]
            if self._row_fmt is not None:
                row_head = f"{self._row_head:{len(row_txt)}s}"
                headings = [row_head] + headings

            write(" ".join(headings) + "\n")
            write(" ".join("-" * len(h) for h in headings) + "\n")
            self._first = False

        if self._row_fmt is not None:
            elem = [row_txt] + elem
        write(" ".join(elem) + "\n")
        self._row = row + 1



"""This module implements functionality related to property handling"""

from collections.abc import Mapping, Iterator
from typing import Self

from ..utilities.quantity import Quantity
from ..utilities.errors import DataFlowError
from ..utilities.types import Map, MutMap


class PropertyHandler(Mapping[str, Quantity]):
    """This class, being instantiated as the :attr:`simu.Model.properties`
    attribute, allows to declare and define process properties."""

    def __init__(self):
        self.__model_name = "N/A"
        self.__props: MutMap[Quantity] = {}
        self.__declared: MutMap[Quantity] = {}

    def __iter__(self) -> Iterator[str]:
        return iter(self.__props)

    def __len__(self) -> int:
        return len(self.__props)

    @staticmethod
    def __raise(name: str, msg: str):
        raise DataFlowError(f"Property '{name}' {msg}")

    def __setitem__(self, name: str, quantity: Quantity):
        """Via this operator, a calculated property is defined as a
        model result."""

        # TODO: is it not ok to define properties that are not declared?
        #  if it is, remove the commended code
        # if name not in self.__declared:
        #     self.__raise(name,  "is not declared")
        if name in self.__props:
            self.__raise(name,  "is already defined")

        # check unit compatibility -> throw exception on demand
        try:
            quantity.to(self.__declared[name])
        except KeyError:
            pass  # property wasn't declared, no problem

        self.__props[name] = quantity

    def __getitem__(self, name: str) -> Quantity:
        """Return a property as it has been defined via the ``__setitem__``
        operator (``[]``) before."""
        return self.__props[name]

    def __enter__(self) -> Self:
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    def declare(self, name: str, unit: str) -> None:
        """This method declares a property to be provided by the model."""
        if name in self.__declared:
            self.__raise(name,  "is already declared")
        self.__declared[name] = Quantity(unit)  # also assure the unit is valid

    def check_complete(self) -> None:
        """Check that all declared properties are defined"""
        if missing := {n for n in self.__declared if n not in self.__props}:
            msg = "The following properties are not defined: " + \
                  ", ".join(missing)
            raise DataFlowError(msg)

    def create_proxy(self) -> "PropertyProxy":
        """Create a proxy object for configuration in hierarchy context"""
        return PropertyProxy(self)


class PropertyProxy(Map[Quantity]):
    """This class is instantiated by the parent's :class:`PropertyHandler`
    to handle the property availability to the parent context."""

    def __init__(self, handler: PropertyHandler):
        self.__handler = handler
        self.__finalised = False

    def __getitem__(self, name: str) -> Quantity:
        if self.__finalised:
            return self.__handler[name]
        else:
            raise DataFlowError("Property access before model is finalised")

    def __len__(self) -> int:
        if self.__finalised:
            return len(self.__handler)
        else:
            raise DataFlowError("Property access before model is finalised")

    def __iter__(self) -> Iterator[str]:
        if self.__finalised:
            return iter(self.__handler)
        else:
            raise DataFlowError("Property access before model is finalised")

    def finalise(self) -> None:
        """Signal that the model is done with its declaration, and properties
        are now available for the target context."""
        self.__handler.check_complete()
        self.__finalised = True



"""This module defines a parser for formulae that can be used to establish
generic element balances, and to calculate molecular weights."""

from re import compile as re_compile
from yaml import safe_load

from .structures import MCounter
from .quantity import Quantity
from ..data import DATA_DIR


class FormulaParser:
    """This class implements the functionality to analyse chemical sum
    formulae. The atomic composition and molecular weight can be obtained.
    """
    VALID_REG = re_compile(r"^[A-Za-z0-9.·\[{(<>)}\]=≡|\-+]+(:\d+[+-])?$")
    EL_REG = re_compile(r"([A-Z][a-z]*|[(])")
    NUM_REG = re_compile(r"\d+(\.\d*)?")
    CRYSTAL_REG = re_compile(r"·(\d+)?([^·]+)")
    STRUC_REG = re_compile(r"[=\-<>≡|+]")
    CHARGE_REG = re_compile(r":\d+[+-]$")

    def __init__(self):
        filename = DATA_DIR / "atomic_weights.yml"
        with open(filename, encoding="utf-8") as file:
            data = safe_load(file)
        self._atomic_weights = {
            sym: Quantity(mw, "g/mol")
            for sym, [_, mw] in data.items()
        }
        self._element_counters = {
            s: MCounter({s: 1})
            for s in self._atomic_weights
        }

    @property
    def atomic_weights(self) -> dict[str, Quantity]:
        """Return a dictionary with all elements mapped to their molecular
        weight"""
        return self._atomic_weights

    def parse(self, formula: str) -> dict[str, int]:
        """Parse a formula and return a Counter object with the atomic symbols
        as keys and the number of occurances as values:

        Plain formulae:
            >>> parser = FormulaParser()
            >>> parser.parse("H3PO4")
            MCounter({'O': 4, 'H': 3, 'P': 1})
            >>> parser.parse("KMnO4")
            MCounter({'O': 4, 'K': 1, 'Mn': 1})
            >>> parser.parse("FISH")
            MCounter({'F': 1, 'I': 1, 'S': 1, 'H': 1})

        With parantheses:
            >>> parser.parse("(NH4)2HPO4")
            MCounter({'H': 9, 'O': 4, 'N': 2, 'P': 1})

        With structure:
            >>> parser.parse("CH3-(CH2)3-CH=O>")
            MCounter({'H': 10, 'C': 5, 'O': 1})
            >>> parser.parse("|N≡N|")
            MCounter({'N': 2})
            >>> parser.parse("<O=O>")
            MCounter({'O': 2})

        With charge:
            >>> parser.parse("SO4:2-")
            MCounter({'O': 4, 'S': 1})

        With a complex:
            >>> parser.parse("Na(UO2)3[Zn(H2O)6](CH3CO2)9")
            MCounter({'H': 39, 'O': 30, 'C': 18, 'U': 3, 'Na': 1, 'Zn': 1})

        With crystal water:
            >>> parser.parse("CuSO4·5H2O")
            MCounter({'H': 10, 'O': 9, 'Cu': 1, 'S': 1})

        With crystal water and another solvent:
            >>> parser.parse("CuSO4·3H2O·2(CH3)-COOH")
            MCounter({'H': 14, 'O': 11, 'C': 4, 'Cu': 1, 'S': 1})

        .. warning::

            This code uses the evil ``eval`` function that yields potential and
            very real security risks in certain contexts. Here it is considered
            safe for the following reasons.

            - the expression must come from a valid formula and is modified
                into the expression itself.
            - global variable dictionary is set to be empty, and local
                variables are only the element counter objects
            - SiMu is running on the PC of the user. Even if the high level
                models are published via network, the formulae are not part of
                the remote user interface.

        However, If this code was used to offer a web-based formula parser,
        some evil genius might in theory find a way to harm the host system
        this way.
        """

        def repl_el(arg):
            """Replace element ``El`` with ``+El``"""
            return f"+{arg.group(0)}"

        def repl_fac(arg):
            """Replace coefficient ``3`` with ``*3``"""
            return f"*{arg.group(0)}"

        if self.VALID_REG.match(formula) is None:
            raise ValueError(f"Invalid syntax of formula: {formula}")

        # Let's turn the formula into an equation with elements as symbols:
        # E.g. (NH4)2SO4 to (+N+H*4)*2+S+O*4
        # This is done by prepending '+' to each element and '*' to each
        # coefficient.

        f_mod = self.CHARGE_REG.sub("", formula)
        for c in "{( [( }) ])".split():
            f_mod =  f_mod.replace(c[0], c[1])
        f_mod = self.CRYSTAL_REG.sub(r"(\2)\1", f_mod)
        f_mod = self.STRUC_REG.sub("", f_mod)
        f_mod = self.EL_REG.sub(repl_el, f_mod)
        f_mod = self.NUM_REG.sub(repl_fac, f_mod)

        # now it can just be evaluated.
        scope = self._element_counters
        try:
            return eval(f_mod, {}, scope)  # pylint: disable=eval-used
        except NameError as err:
            raise ValueError(
                f"Formula contains invalid element: {err}") from None

    def molecular_weight(self, formula: str) -> Quantity:
        """Return the molecular weight for a given formula. The atomic weights
        are originally obtained from :cite:p:`CoplenShrestha_2016`.

            >>> parser = FormulaParser()
            >>> mw = parser.molecular_weight("CH3-(CH2)24-CH3")
            >>> print(f"{mw:~.2f}")
            366.72 g / mol

        """
        elements = self.parse(formula)
        weights = self.atomic_weights
        w_0 = Quantity(0, "g/mol")
        return sum((weights[sym] * nu for sym, nu in elements.items()), w_0)

    def charge(self, formula: str) -> Quantity:
        """Return the charge associated to the given formula.

            >>> parser = FormulaParser()
            >>> for n in "H2SO4 SO4:2- Al:3+ S6:12-".split():
            ...     print(f"{n}: {parser.charge(n):~}")
            H2SO4: 0 e / mol
            SO4:2-: -2 e / mol
            Al:3+: 3 e / mol
            S6:12-: -12 e / mol
        """
        if self.VALID_REG.match(formula) is None:
            raise ValueError(f"Invalid syntax of formula: {formula}")

        match = self.CHARGE_REG.search(formula)
        if match is None:
            return Quantity(0, "e / mol")
        raw = match.group(0)
        return Quantity(int(f"{raw[-1]}{raw[1:-1]}"), "e / mol")



# stdlib modules
from abc import ABC, abstractmethod

# external modules
from casadi import SX
from numpy import roots

# internal modules
from simu import (
    InitialState, base_magnitude, jacobian, log, qsum, R_GAS,
    ThermoContribution, Quantity
)


class RedlichKwongEOS(ThermoContribution, ABC):
    r"""This contribution implements a general Redlich-Kwong equation of state
    with Peneloux volume translation:

    .. math::

        p = \frac{N\,R\,T}{V - B + C} - \frac{A}{(V + C)\,(V + B + C)}

    The following properties need to be provided by upstream contributions:

    ======== ========= ======== ==========
    Property Symbol    UOM
    ======== ========= ======== ==========
    ceos_a   :math:`A` J * |m3|
    ceos_b   :math:`B` |m3|
    ceos_c   :math:`C` |m3|     (optional)
    ======== ========= ======== ==========

    Care is to be taken when utilising a temperature-dependent :math:`C`
    contribution, as doing so can have significant effects on the calorimetric
    properties. If ``ceos_c`` is not provided, the contribution is assumed
    zero.

    As such, there are no further model parameters to be provided at this
    point. The residual Helmholtz function is

    .. math::
        A^\mathrm{res} = \int\limits_V^\infty
           p - \frac{N\,R\,T}{V} \mathrm{d}V
         = N\,R\,T\,\ln \frac{V}{V + C - B} +
           \frac{A}{B}\,\ln \frac{V + C}{V + C + B}

    The explicitly implemented temperature derivative is

    .. math::

        -S^\mathrm{res} &= N\,R\,\left [
            \ln\frac{V}{V - B + C} + T\,\frac{B_T - C_T}{V - B + C}
            \right ]\\& +
            \frac1B\left (A_T - \frac{A}{B}\,B_T\right )\,
            \ln \frac{V + C}{V + B + C} +
            \frac{A}{B}\left  [
                \frac{C_T}{V + C} - \frac{B_T + C_T}{V + B + C}
            \right ]

    The volume derivative is the negative residual pressure:

    .. math::

        -p^\mathrm{res} =
          N\,R\,T\, \left [ \frac1V - \frac1{V - B + C}\right ] +
          \frac{A}{(V + C)\,(V + B + C)}

    The derivative with respect to molar quantities is

    .. math::

        \mu_i^\mathrm{res} &= R\,T\,\left [
            \ln\frac{V}{V - B + C} + N\,\frac{B_{n,i} - C_{n,i}}{V - B + C}
            \right ]\\& +
            \frac1B\left (A_{n,i} - \frac{A}{B}\,B_{n,i}\right )\,
            \ln \frac{V + C}{V + B + C} +
            \frac{A}{B}\left  [
                \frac{C_{n,i}}{V + C} - \frac{B_{n,i} + C_{n,i}}{V + B + C}
            \right ]

    The contribution updates are

    .. math::

        S &\leftarrow S + S^\mathrm{res}\\
        p &\leftarrow p + p^\mathrm{res}\\
        \mu_i &\leftarrow \mu_i + \mu_i^\mathrm{res}\\

    .. important::

        The original publication on the volume correction of EOS
        :cite:p:`Peneloux82` suggests to shift both V and B, not only V. The
        derivation of this proposal is not given, and a numerical experiment
        shows that doing so drastically impacts the equilibrium and the density
        predictions in a negative way. We therefore stick to the approach taken
        by :cite:p:`AspenTech2001`, namely a pure volume translation. For this
        translation, it can easily be shown, as by :cite:p:`Peneloux82` that
        equilibrium conditions are not affected:

        Consider the chemical potential derived from the Helmholtz function

        .. math::
            A = A(T, V, \vec n)\quad\Rightarrow\quad
            \vec \mu = \left . \frac{\partial A}{\partial \vec n}
              \right |_{T, V}

        If we translate :math:`\tilde V = V - \vec c\cdot\vec n`, we can derive

        .. math::
            \tilde A = \tilde A(T, \tilde V, \vec n)\quad\Rightarrow\quad
             \vec{\tilde \mu} = \left . \frac{\partial \tilde A}{\partial \vec n}
              \right |_{T, \tilde V}
              = \left . \frac{\partial \tilde A}{\partial \vec n}
              \right |_{T, V} +
              \left . \frac{\partial \tilde A}{\partial \tilde V}
              \right |_{T, n}\cdot
              \left . \frac{\partial \tilde V}{\partial \vec n}
                \right |_{T}
              =  \vec \mu - p\,\vec c

        Hence at given temperature and pressure, the chemical potential only
        shifts by a constant, not affecting any equilibrium constraints.
    """

    provides = ["_VCB", "_VCB_x", "_p_x", "_p_V", "_p_V_x"]

    def define(self, res):
        ab_names = ["_ceos_a", "_ceos_b"]
        T, V, n, A, B = [res[i] for i in ["T", "V", "n"] + ab_names]

        for i in ab_names:
            res[f"{i}_T"] = jacobian(res[i], T)
            res[f"{i}_n"] = jacobian(res[i], n).T  # jacobian transposes
        A_t, B_t = [res[f"{i}_T"] for i in ab_names]
        A_n, B_n = [res[f"{i}_n"] for i in ab_names]

        if (c_name := "_ceos_c") in res:  # C is optional
            C = res[c_name]
            C_t = res[f"{c_name}_T"] = jacobian(res[c_name], T)
            C_n = res[f"{c_name}_n"] = jacobian(res[c_name], n).T  # jacobian transposes
        else:
            C = res[c_name] = Quantity(SX(1, 1), "m**3/mol") * n.units
            C_t = res[f"{c_name}_T"] = Quantity(SX(1, 1), "m**3/mol/K") * n
            C_n = res[f"{c_name}_n"] = Quantity(SX(*n.shape), "m**3/mol")

        # common terms
        N = qsum(n)
        NR, RT = N * R_GAS, T * R_GAS
        VC = V + C
        VmBC, VpBC = VC - B, VC + B
        AB = A / B

        # entropy contribution
        m_dS = NR * (log(V / VmBC) + T * (B_t - C_t) / VmBC)
        m_dS += (A_t - AB * B_t) / B * log(VC / VpBC)
        m_dS += AB * (C_t / VC - (B_t + C_t) / VpBC)
        res["S"] -= m_dS

        # pressure contribution
        res["p"] -= NR * T * (1 / V - 1 / VmBC) + A / (VC * VpBC)

        # chemical potential contribution
        dmu = RT * (log(V / VmBC) + N * (B_n - C_n) / VmBC)
        dmu += (A_n - AB * B_n) / B * log(VC / VpBC)
        dmu += AB * (C_n / VC - (B_n + C_n) / VpBC)
        res["mu"] += dmu

        self.add_bound("RK_VmBC", VmBC)  # V - B + C > 0
        self.add_bound("V", V)
        self.add_bound("dp_dV", -jacobian(res["p"], V))  # dp/dv < 0
        self.add_bound("p", res["p"])

    @staticmethod
    def find_zeros(state: InitialState, properties):
        """Given conditions and properties (A, B, C contribution of RK-EOS),
        calculate the compressibility factor roots analytically and return
        the vector quantity of volumes based on these"""
        T, p, n = state.temperature, state.pressure, state.mol_vector
        A, B, C = [properties[f"_ceos_{i}"] for i in "abc"]
        NRT = sum(n) * R_GAS * T
        alpha = float(A * p / (NRT * NRT))  # must be dimensionless
        beta = float(B * p / NRT)  # dito
        zeros = roots([1, -1, alpha - beta * (1 + beta), -alpha * beta])
        zeros = zeros[abs(zeros.imag) < 1e-7 * abs(zeros)]
        return zeros[zeros > beta].real * NRT / p - C


    @abstractmethod
    def initial_state(self, state, properties):
        """Force implementation of this method"""
        ...


class RedlichKwongEOSLiquid(RedlichKwongEOS):
    """As a subclass of
    :class:`~simu.app.thermo.contributions.cubic.rk.RedlichKwongEOS`, this
    entity specialises on describing liquid (and super-critical) phases.
    The distinct element is the initialisation.
    """

    def initial_state(self, state, properties):
        volume = min(self.find_zeros(state, properties))
        return ([base_magnitude(state.temperature), base_magnitude(volume)] +
                list(base_magnitude(state.mol_vector)))


class RedlichKwongEOSGas(RedlichKwongEOS):
    """As a subclass of
    :class:`~simu.app.thermo.contributions.cubic.rk.RedlichKwongEOS`, this
    entity specialises on describing gas (and super-critical) phases. The
    distinct elements is the initialisation."""

    def initial_state(self, state, properties):
        zeros = self.find_zeros(state, properties)
        volume = max(zeros)
        return ([base_magnitude(state.temperature), base_magnitude(volume)] +
                list(base_magnitude(state.mol_vector)))


class RedlichKwongAFunction(ThermoContribution):
    r"""Given critical temperature ``T_c`` (:math:`T_{c,i}`) and pressure
    ``p_c`` (:math:`p_{c,i}`), this contribution scales the
    :math:`\alpha`-function ``alpha`` (:math:`\alpha_i`) to define the
    :math:`a`-contribution ``rk_a_i`` (:math:`a_i`) for the individual species.
    It is

    .. math::

        a_i = \alpha_i\,\Omega_a\,\frac{R^2\,T_{c,i}^2}{p_{c,i}}
        \quad\text{with}\quad
        \Omega_a = \frac19\,(2^{1/3} - 1)^{-1}
    """

    provides = ["_ceos_a_i"]

    def define(self, res):
        omega_r2 = R_GAS * R_GAS / (9 * (2**(1 / 3) - 1))
        alpha, T_c, p_c = [res[i] for i in "_alpha _T_c _p_c".split()]
        res["_ceos_a_i"] = omega_r2 * alpha * (T_c * T_c) / p_c


class RedlichKwongBFunction(ThermoContribution):
    r"""Given critical temperature ``T_c`` (:math:`T_{c,i}`) and pressure
    ``p_c`` (:math:`p_{c,i}`), this contribution calculates the
    :math:`b`-contribution ``ceos_b_i`` for the individual species. It is

    .. math::

        b_i = \Omega_b\,\frac{R\,T_{c,i}}{p_{c,i}}
        \quad\text{with}\quad
        \Omega_b = \frac13\,(2^{1/3} - 1)
    """

    provides = ["_ceos_b_i"]

    def define(self, res):
        omega_r = R_GAS * (2**(1 / 3) - 1) / 3
        T_c, p_c = [res[i] for i in "_T_c _p_c".split()]
        res["_ceos_b_i"] = omega_r * T_c / p_c


class RedlichKwongMFactor(ThermoContribution):
    r"""This contribution calculates the Redlich Kwong m-factor that is used
    in various alpha-functions. Based on provided acentric factors ``omega``
    (:math:`\omega_i`), it calculates ``m_factor`` (:math:`m_i`) as

    .. math::

        m_i = 0.48508 + (1.55171 - 0.15613\,\omega_i)\,\omega_i
    """

    provides = ["_m_factor"]

    def define(self, res):
        omega = res["_omega"]
        m = 0.48508 + (1.55171 - 0.15613 * omega) * omega
        res["_m_factor"] = Quantity(m)



"""This module defines classes and functionality around ``pint`` quantities.
These quantities can be symbolic (hosting ``casadi.SX`` as magnitudes), or
numeric."""

# stdlib modules
from typing import Union, Self

# external modules
# need to import entire casadi module to distinguish functions of same name
import casadi as cas
from numpy import squeeze
from pint import UnitRegistry, set_application_registry, Unit
from pint.util import UnitsContainer
from pint.errors import DimensionalityError

# internal modules
from . import flatten_dictionary, unflatten_dictionary
from .types import NestedMap, Map, MutMap
from ..data import DATA_DIR


def _create_registry() -> UnitRegistry:
    """Instantiate the application's unit registry"""
    reg = UnitRegistry(autoconvert_offset_to_baseunit=True)
    set_application_registry(reg)
    unit_file = DATA_DIR / "uom_definitions.txt"
    reg.load_definitions(unit_file)
    return reg


unit_registry = _create_registry()
_Q = unit_registry.Quantity
del _create_registry

__DERIVED_UNITS = \
    "joule watt newton pascal coulomb volt farad ohm " \
    "siemens weber tesla henry lumen lux".split()



class Quantity(_Q):
    """Proper quantity base-class for sub-classing.

    Being a subclass of ``pint.Quantity``, this class only really adds the
    ``__json__`` method to return its json representation.

    The constructor is used as for ``pint.Quantity``.
    """

    def __new__(cls, *args, **kwargs):
        obj = super().__new__(_Q, *args, **kwargs)
        obj.__class__ = cls
        return obj

    def __json__(self) -> str:
        """Custom method to export to json for testing and serialisation"""
        try:
            return f"{self:~.16g}"
        except TypeError:  # magnitude cannot be formatted, probably symbol
            return f"{self:~}"

    # operators needed for proper type checking in Pycharm (so long at least)
    def __add__(self, other) -> Self:
        return super().__add__(other)

    def __radd__(self, other) -> Self:
        return super().__radd__(other)

    def __sub__(self, other) -> Self:
        return super().__sub__(other)

    def __rsub__(self, other) -> Self:
        return super().__rsub__(other)

    def __mul__(self, other) -> Self:
        return super().__mul__(other)

    def __rmul__(self, other) -> Self:
        return super().__rmul__(other)

    def __truediv__(self, other) -> Self:
        return super().__truediv__(other)

    def __rtruediv__(self, other) -> Self:
        return super().__rtruediv__(other)

    def __pow__(self, other) -> Self:
        return super().__pow__(other)

    def __rpow__(self, other) -> Self:
        return super().__rpow__(other)


# write back class, so it's used as result of quantity operations
unit_registry.Quantity = Quantity


class SymbolQuantity(Quantity):
    """A quantity class specialised to host casadi symbols (SX) of in
    particular, but not necessarily, independent variables."""

    def __new__(cls, *args, **kwargs):
        """Really generate an object of type ``Quantity``. This is just a
        hacky way to specialise the constructor, due to the way the base-class
        is implemented. The arguments are:

        - ``name`` (str): The name of the ``casadi.SX`` symbol
        - ``units`` (str): The unit of measurement conform to ``pint`` units
        - ``sub_keys`` (Iterator[str]): If not ``None``, defines a vectorial
          quantity with given sub-keys.

            >>> s = SymbolQuantity("speed", "m/s")
            >>> print(f"{s:~}")
            speed m / s

            >>> v = SymbolQuantity("vel", "m/s", "xyz")
            >>> print(f"{v:~}")
            [vel.x, vel.y, vel.z] m / s
        """

        def attributes(name: str, units: str, sub_keys: list[str] = None):
            """Turn the constructor arguments into arguments for the
            baseclass"""
            if sub_keys is None:
                magnitude = cas.SX.sym(name)
            elif isinstance(sub_keys, int):
                magnitude = cas.SX.sym(name, sub_keys)
            else:
                magnitude = [cas.SX.sym(f"{name}.{s}") for s in sub_keys]
                magnitude = cas.vertcat(*magnitude)
            return magnitude, units

        return super().__new__(Quantity, *attributes(*args, **kwargs))

    def __json__(self) -> str:
        """Custom method to export to json for testing"""
        return f"{str(self.magnitude)}{self.units:~}"


# redefine Jacobian to propagate pint units
def jacobian(dependent: Quantity, independent: Quantity) -> Quantity:
    """Calculate the casadi Jacobian and reattach the units of measurements.

    >>> x = SymbolQuantity("x", "m")
    >>> y = (x * x) / 2
    >>> dy_dx = jacobian(y, x)
    >>> print(f"{dy_dx:~}")
    x m
    """
    magnitude = cas.jacobian(dependent.magnitude, independent.magnitude)
    return Quantity(cas.simplify(magnitude),
                    dependent.units / independent.units)


def qsum(quantity: Quantity) -> Quantity:
    """Sum a symbol vector quantity same was a ``casadi.sum1``, considering
    units of measurements. This function only applies to quantity objects with
    ``casadi.SX`` objects as magnitudes.

    .. note::

        This function sums the elements of the single, but vectorial argument,
        and is in that different from the builtin ``sum`` function that sums
        over an iterator of objects.
    """
    return Quantity(cas.sum1(quantity.magnitude), quantity.units)


def qsqrt(quantity: Quantity) -> Quantity:
    """Determine square root of symbolic quantity, considering units of
    measurement"""
    return Quantity(cas.sqrt(quantity.magnitude), quantity.units**0.5)


def qpow(base: Quantity, exponent: Quantity) -> Quantity:
    """Determine power of ``base`` to ``exponent``, considering units of
    measurements. Both arguments must be dimensionless. For the special case
    that ``exponent`` is a constant scalar (not a symbol), use the ``**``
    operator
    """
    base = base.to_base_units()  # else e.g. cm / m is dimensionless
    exponent = exponent.to_base_units()
    if not base.dimensionless:
        raise DimensionalityError(base.units, "dimensionless")
    if not exponent.dimensionless:
        raise DimensionalityError(exponent.units, "dimensionless")
    return Quantity(base.magnitude**exponent.magnitude)


def conditional(condition: cas.SX, negative: Quantity,
                positive: Quantity) -> Quantity:
    """Element-wise branching into the positive and negative branch
    depending on the condition and considering the units of measurements.

        >>> x = SymbolQuantity("x", "m")
        >>> y = conditional(x > 0, -x, x)  # abs function
        >>> print(y)
        @1=0, @2=((@1<x)==@1), ((@2?(-x):0)+((!@2)?x:0)) meter

    As non-recommended as it is to use this function exessively, the resulting
    ``casadi`` expression is indeed overcomplicated and could simplify to ::

        (x>0)?(x):(-x) meter

    .. note::

        You cannot just code ``x if x > 0 else -x``, as this would not allow
        ``casadi`` branching dynamically dependent on the value of ``x``
        later-on.
    """
    units = positive.units
    negative = negative.to(units)  # convert to same unit
    tup_branches = [condition, negative.magnitude, positive.magnitude]
    tup = zip(*map(cas.vertsplit, tup_branches))

    magnitude = cas.vertcat(*[cas.conditional(c, [n], p) for c, n, p in tup])
    return Quantity(magnitude, units)


def qvertcat(*quantities: Quantity) -> Quantity:
    """Concatenate a bunch of scalar symbolic quantities with compatible units
    to a vector quantity"""
    units = quantities[0].units
    magnitudes = [q.to(units).magnitude for q in quantities]
    return Quantity(cas.vertcat(*magnitudes), units)


def base_unit(unit: str) -> str:
    """Create the base unit of given unit.

    >>> print(base_unit("light_year"))
    m
    >>> print(base_unit("week"))
    s
    """
    if unit == "":
        unit = "dimensionless"
    base = unit_registry.Quantity(unit).to_base_units().units
    return f"{base:~}"


def simplify_quantity(quantity: Quantity) -> Quantity:
    """Try to convert the unit into a more compact notation, allowing
    derived units, such as Watt, Joule and Pascal to be used.

    Examples:

    >>> q = Quantity(1.0, "kg * m**2 / mol / s**2")
    >>> print(f"{simplify_quantity(q):~}")
    1.0 J / mol

    >>> q = Quantity(1.0, "kg / K / s**3")
    >>> print(f"{simplify_quantity(q):~}")
    1.0 W / K / m ** 2

    >>> q = Quantity(1.0, "kg / m**2 / s**2")
    >>> print(f"{simplify_quantity(q):~}")
    1.0 Pa / m

    >>> q = Quantity(1.0, "kg * m**2 / s**3 / A**2")
    >>> print(f"{simplify_quantity(q):~}")
    1.0 Ω

    >>> q = Quantity(1.0, "m**3 / s")
    >>> print(f"{simplify_quantity(q):~}")
    1.0 m ** 3 / s
    """
    def weight_factors(dimensionality: UnitsContainer) -> float:
        """Count the factors, but weight an exponent unequal unity less hard
        than a completely new factor
         """
        return sum(1 + abs(e) for e in dimensionality.values())

    quantity = quantity.to_base_units()
    count = weight_factors(quantity.dimensionality)
    for unit in __DERIVED_UNITS:
        trial = Quantity(1, unit)
        candidate = (quantity / trial).to_base_units()
        new_count = weight_factors(candidate.dimensionality)
        if new_count < count:
                quantity = candidate * trial
                count = new_count
    return quantity


def base_magnitude(quantity: Quantity) -> Union[float, "cas.SX"]:
    """Return the magnitude of the quantity in base units. This works for
    symbolic and numeric quantities, and for scalars and vectors

    >>> base_magnitude(Quantity("1 km"))
    1000.0
    >>> base_magnitude(Quantity("20 degC"))
    293.15
    >>> int(base_magnitude(Quantity("1 barg")))
    201325
    >>> base_magnitude(Quantity("speed_of_light"))
    299792458.0

    The base units are likely the SI unit system, but code shall not rely on
    this fact - only that it is a cosistent (and offset-free) unit system.

    .. note::

        "**Use SI or I will BTU you with my feet!**"

        I saw this sentence once on the T-shirt of a nerd.
        It turns out it was a mirror.
    """
    return quantity.to_base_units().magnitude


def extract_units_dictionary(structure: NestedMap[Quantity] | Quantity) \
        -> NestedMap[str] | str:
    """Based on a nested dictionary of Quantities, create a new nested
    dictionaries with only the (base) units of measurement

    >>> d = {"a": {"b": Quantity("1 m"), "c": Quantity("1 min")}}
    >>> extract_units_dictionary(d)
    {'a': {'b': 'm', 'c': 's'}}
    """
    try:
        items = structure.items()
    except AttributeError:
        return f"{simplify_quantity(structure).units:~}"
    return {key: extract_units_dictionary(value) for key, value in items}


class QFunction:
    """Wrapper around ``casadi.Function`` to consider units of measurements.
    This derived function object is defined by a dictionary of arguments and
    results, both of which are ``Quantity`` instances with casadi symbols as
    magnitudes. These symbols are independent symbols for ``args`` and derived
    symbols for ``results``.

    The function object is then called as well with a dictionary of
    ``Quantity`` values. The units of measurement must be consistent, and a
    conversion is done to the initially defined units for the individual
    arguments. The result is given back as a dictionary of ``Quantity`` values
    in the same units as initially defined.
    """

    def __init__(self, args: NestedMap[Quantity], results: NestedMap[Quantity],
                 func_name: str = "f"):
        args_flat = flatten_dictionary(args)
        results_flat = flatten_dictionary(results)
        arg_sym = cas.vertcat(*[v.magnitude for v in args_flat.values()])

        self.__res_shapes = {}
        res_sym = []
        for key, value in results_flat.items():
            mag = value.magnitude
            shape = mag.shape
            self.__res_shapes[key] = mag.shape
            res_sym.append(cas.reshape(mag, shape[0] * shape[1], 1))
        res_sym = cas.vertcat(*res_sym)

        self.arg_units = {k: v.units for k, v in args_flat.items()}
        self.res_units = {k: v.units for k, v in results_flat.items()}
        self.func = cas.Function(func_name, [arg_sym], [res_sym],
                                 ["x"], ["y"])

    def __call__(self, args: NestedMap[Quantity],
                 squeeze_results: bool = True) -> NestedMap[Quantity]:
        """Call operator for the function object, as described above.

        :param args: The arguments of the function
        :param squeeze_results: The underlying `Casadi`_ function returns
        the result objects in 2D shapes. With this parameter being ``True``
        (default), ``numpy.squeeze`` is applied to all results, to omit
        dimensions that are of size one.
        """
        args_flat = cas.vertcat(*[
            value.to(self.arg_units[key]).magnitude
            for key, value in flatten_dictionary(args).items()
        ])
        result = self.func(args_flat)  # calling Casadi function
        result = self.__unpack(result)
        if squeeze_results:
            result = {k: squeeze(v) for k, v in result.items()}
        result = {k: simplify_quantity(Quantity(v, self.res_units[k]))
                  for k, v in result.items()}
        return unflatten_dictionary(result)

    def __unpack(self, raw_result: cas.SX) -> Map[cas.SX]:
        result: MutMap[cas.SX] = {}
        idx = 0
        for key, shape in self.__res_shapes.items():
            size = shape[0] * shape[1]
            result[key] = raw_result[idx:idx + size].reshape(shape)
            idx += size
        return result

    @property
    def result_structure(self) -> NestedMap[str]:
        """Return the result structure as a nested dictionary, only including
        the units of measurements as values of end nodes"""
        simplify = self.__simplify_unit
        units = {k: f"{simplify(v):~}" for k, v in self.res_units.items()}
        return unflatten_dictionary(units)

    @property
    def arg_structure(self) -> NestedMap[str]:
        """Return the argument structure as a nested dictionary, only including
        the units of measurements as values of end nodes"""
        simplify = self.__simplify_unit
        units = {k: f"{simplify(v):~}" for k, v in self.arg_units.items()}
        return unflatten_dictionary(units)

    @staticmethod
    def __simplify_unit(unit: Unit) -> Unit:
        return simplify_quantity(Quantity(1, unit)).units



"""This module handles functionality concerning model hierarchy."""
from typing import TYPE_CHECKING, Type, Any
from collections.abc import Mapping, Iterator

from ..utilities.types import Map, MutMap
from ..utilities.errors import DataFlowError

if TYPE_CHECKING:  # avoid circular dependencies just for typing
    from .base import Model, ModelProxy


class HierarchyHandler(Map["ModelProxy"]):
    """This class, being instantiated as the :attr:`simu.Model.hierarchy`
    attribute, allows to define child models in a hierarchy context."""

    def __init__(self, model: "Model"):
        self.model = model
        self.__children: MutMap["ModelProxy"] = {}
        self.__declared: MutMap[Type["Model"]] = {}

    def __len__(self) -> int:
        return len(self.__children)

    def __iter__(self) -> Iterator[str]:
        return iter(self.__children)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return

    def declare(self, name: str, model_cls: Type["Model"]) -> None:
        """Declare a sub-model in the interface, and by that

        a) Demand that it will be instantiated, and
        b) Make it available in the hierarchy proxy for browsing from parent
           context.
        """
        if name in self.__declared:
            raise KeyError(f"Child model '{name}' already declared")
        self.__declared[name] = model_cls

    def add(self, name: str, model_cls: Type["Model"],
            *args: Any, **kwargs: Any) -> "ModelProxy":
        """Add an instance of the class ``model_cls`` as child to the current
        (parent) context. A :class:`~simu.core.model.base.ModelProxy` object is
        created, registered, and returned."""
        if name in self.__children:
            raise KeyError(f"Child model '{name}' already exists")
        if (name in self.__declared and
                not issubclass(model_cls, self.__declared[name])):
            cls_name = self.__declared[name].__name__
            msg = f"Model '{name}' is not a subclass of {cls_name} as declared"
            raise ValueError(msg)

        instance = model_cls(*args, **kwargs).create_proxy(name)
        self.__children[name] = instance
        return instance

    @property
    def declared(self) -> Map[Type["Model"]]:
        """Dictionary of declared sub-models"""
        return self.__declared

    def __getitem__(self, name: str):
        """Re-obtain the proxy of named module, avoiding the need to keep a
        holding variable in the client scope code."""
        return self.__children[name]

    def create_proxy(self) -> "HierarchyProxy":
        """Create a proxy object for configuration in hierarchy context"""
        return HierarchyProxy(self)

    def check_complete(self):
        """Check that all declared child models are defined"""
        if missing := {n for n in self.__declared if n not in self.__children}:
            msg = "The following declared child models are not defined: " + \
                  ", ".join(missing)
            raise DataFlowError(msg)


class HierarchyProxy(Mapping[str, "ModelProxy"]):
    """A wrapper of the HierarchyHandler to grant access to the previously
    declared sub-models."""

    def __init__(self, handler: HierarchyHandler):
        self.handler = handler

    def __getitem__(self, name: str) -> "ModelProxy":
        if name not in self.handler.declared:
            raise KeyError(f"Child model of name 'name' not declared")
        return self.handler[name]

    def __len__(self) -> int:
        return len(self.handler.declared)

    def __iter__(self) -> Iterator[str]:
        return iter(self.handler.declared)

    def finalise(self):
        """Check that the child models declared in the interface are
        implemented"""
        self.handler.check_complete()



from typing import cast
from pytest import fixture, raises

from simu import NumericHandler, SimulationSolver, Quantity
from simu.core.utilities.residual import ResidualHandler
from simu.core.utilities.errors import NonSquareSystem
from simu.core.utilities.qstructures import quantity_dict_to_strings
from simu.core.utilities import assert_reproduction
from simu.examples.material_model import Source


def test_instantiate():
    numeric = NumericHandler(Source.top())
    _ = SimulationSolver(numeric)


def test_solve(sim_result):
    assert len(sim_result.iterations) < 5


def test_solve_res_small(sim_result):
    res = sim_result.properties[NumericHandler.VECTORS][NumericHandler.RES_VEC]
    assert (abs(res.m_as("")) < 1).all()


def test_reproduce_result(sim_result):
    n = sim_result.properties["thermo_props"]["source"]["n"]["Methane"]
    assert abs(n.m_as("mol/s") - 0.112054293180843) < 1e-7


def test_non_square():
    """add another residual and thus make system non-square"""
    model = Source().create_proxy().finalise()
    residuals = cast(ResidualHandler, model.residuals) # need to add one more
    residuals.add("Q", model.residuals["T"].value, "K")
    numeric = NumericHandler(model)
    with raises(NonSquareSystem):
        _ = SimulationSolver(numeric)


def test_model_parameters():
    numeric = NumericHandler(Source.top())
    solver = SimulationSolver(numeric, output="None")
    param = quantity_dict_to_strings(solver.model_parameters["model_params"])
    assert_reproduction(param)


def test_invalid_option_constructor():
    numeric = NumericHandler(Source.top())
    with raises(ValueError):
        _ = SimulationSolver(numeric, max_iter=-3)


def test_valid_option_set_option():
    numeric = NumericHandler(Source.top())
    solver = SimulationSolver(numeric)
    solver.set_option("max_iter", 20)


def test_invalid_option_set_option():
    numeric = NumericHandler(Source.top())
    solver = SimulationSolver(numeric)
    with raises(ValueError):
        solver.set_option("max_iter", -3)


def test_change_parameters():
    numeric = NumericHandler(Source.top())
    solver = SimulationSolver(numeric, output="None")
    param = solver.model_parameters["model_params"]
    param["p"] = Quantity(2, "MPa")
    res = solver.solve(max_iter=5)
    p = res.properties["thermo_props"]["source"]["p"]
    assert abs(p - param["p"]) < Quantity(1e-7, "bar")


    # The following code can be start of providing a report unit system
    # from simu import flatten_dictionary
    # props = flatten_dictionary(res.result["thermo_props"])
    # default_units = {
    #     "[temperature]": "degC",
    #     "[length]": "m",
    #     "[volume]": "m^3",
    #     "[length] ** 3 / [time]": "m^3/hr",  # Volumetric flow
    #     "[time]": "s",
    #     "[mass] / [length] / [time] ** 2": "bar"
    # }
    # for k, q in props.items():
    #     try:
    #         unit = default_units[str(q.dimensionality)]
    #     except KeyError:
    #         pass  # print(k, q.dimensionality)
    #     else:
    #         q = q.to(unit)
    #     print(f"{k:20s} {q:.3g~}")


@fixture(scope="module")
def sim_result():
    numeric = NumericHandler(Source.top())
    solver = SimulationSolver(numeric, output="None")
    return solver.solve()



# -*- coding: utf-8 -*-

# internal modules
from simu import jacobian, ThermoContribution


class Derivative(ThermoContribution):
    r"""This auxiliary contribution provides the derivative of an arbitrary
    property with respect to an independent (state) variable. This can for
    instance be used to equip a thermodynamic model with extra temperature
    derivatives for calculating heat capacity or partial molar enthalpy
    as canonical properties.

    The contribution requires an ``option`` dictionary with the following
    entries:

        - ``x``: The name of the independent property :math:`x`
        - ``y``: The name of the dependent property :math:`y`

    The derivative :math:`\partial y/\partial x` will be provided as
    ``f"d{options['y']}_d{options['x']}"``.
    """

    def define(self, res):
        independent = self.options["x"]
        dependent = self.options["y"]
        name = f"d{dependent}_d{independent}"
        if name not in res:
            res[name] = jacobian(res[dependent], res[independent])

        # TODO: what do I do if one or even both of these are vectors?
        #  if one is a vector, I can still use the normal declare_vector_keys,
        #  but if both are vectors, I need to generalize that routine.


"""Simple module to define the version - such that it can be read from other
places as well."""

MAJOR : int = 0  #: Major version
MINOR : int = 4  #: Minor version
FLAG : None | str = None  #: Version flag (None if release, else alpha or beta)
BUILD : int = 1  #: Build flag

VERSION = f"{MAJOR}.{MINOR}"
if FLAG is not None:
    VERSION = f"{VERSION}-{FLAG[0]}{BUILD}"



# -*- coding: utf-8 -*-

# stdlib modules
from copy import copy
from typing import Sequence

# internal modules
from simu import ThermoContribution, R_GAS, log, qsum, base_magnitude, qvertcat
from simu.core.utilities.errors import DimensionalityError


class H0S0ReferenceState(ThermoContribution):
    r"""This contribution defines the reference state based on enthalpy of
    formation and standard entropy. It requires the following parameters:

    =========== ================================== ======================
    Parameter   Description                        Symbol
    =========== ================================== ======================
    ``dh_form`` Molar enthalpy of formation vector :math:`\Delta_f h`
    ``s_0``     Vector of standard entropies       :math:`s_0`
    ``T_ref``   Reference temperature              :math:`T_\mathrm{ref}`
    ``p_ref``   Reference pressure                 :math:`T_\mathrm{ref}`
    =========== ================================== ======================

    Based on this, the reference state is defined in therms of entropy
    and chemical potentials:

    ========= ========================================= ======================
    Property  Description                               Symbol
    ========= ========================================= ======================
    ``S``     Reference state entropy of the system     :math:`S`
    ``mu``    Reference state chemical potential vector :math:`\mu`
    ``T_ref`` Reference temperature (same as parameter) :math:`T_\mathrm{ref}`
    ``p_ref`` Reference pressure (same as parameter)    :math:`p_\mathrm{ref}`
    ========= ========================================= ======================

    The contribution does not support any options.

    .. math::

        S &= \sum_i s_{0,i}\, n_i\\
        \mu_i &= \Delta_f h_i - T\,s_{0,i}
    """

    provides = ["T_ref", "p_ref", "S", "mu"]

    def define(self, res):
        species = self.species

        s_0 = self.par_vector("s_0", species, "J/(mol*K)")
        dh_form = self.par_vector("dh_form", species, "J/mol")
        res["S"] = s_0.T @ res["n"]
        res["mu"] = dh_form - res["T"] * s_0
        res["T_ref"] = self.par_scalar("T_ref", "K")
        res["p_ref"] = self.par_scalar("p_ref", "Pa")

        self.declare_vector_keys("mu")


class LinearHeatCapacity(ThermoContribution):
    r"""This contribution implements a simple heat capacity, being linear
    in temperature. It normally builds on a reference state and yields the
    standard state model. It requires the following parameters:

    ========= ====================================== ===============
    Parameter Description                            Symbol
    ========= ====================================== ===============
    ``cp_a``  Molar heat capacity coefficient vector :math:`c_{p,a}`
    ``cp_b``  Molar heat capacity coefficient vector :math:`c_{p,b}`
    ========= ====================================== ===============

    Heat capacity is here defined as

    .. math::

        \Delta T := T-T_\mathrm{ref}\\
        c_{p,i} = c_{p,a,i} + \Delta T\,c_{p,b,i}

    The contribution does not define new properties, but builds on entropy
    ``S`` (:math:`S`) and chemical potential ``mu`` (:math:`\mu`). With the
    following definitions:

    .. math::

        \Delta h_i &= \left (c_{p,a,i} + \frac12\Delta T\,c_{p,b,i}\right )\,
                      \Delta T\\
        \Delta s_i &= (c_{p,a,i} - c_{p,b,i}\,T_\mathrm{ref})\,
                      \ln \frac{T}{T_\mathrm{ref}} +
                      c_{p,b,i}\,\Delta T\\

    The properties are updated as

    .. math::

        S &\leftarrow S + \sum_i \Delta s_i\,n_i\\
        \mu_i &\leftarrow \mu_i + \Delta h_i - T\,\Delta s_i

    .. todo:
        - use independent T_0 for cp = A  + B * (T - T_0), so it is
          invariant to T_ref. Then
          H = A * (T - T_ref) + B / 2 * (T^2 - T_ref^2 - T_0 * T + T_0 * T_ref)

    The contribution model domain is defined for positive temperatures only,
    due to a logarithmic contribution to entropy.
    """

    def define(self, res):
        T, n = res["T"], res["n"]
        T_ref = res["T_ref"]
        d_T, f_T = T - T_ref, T / T_ref
        cp_a = self.par_vector("cp_a", self.species, "J/(mol*K)")
        cp_b = self.par_vector("cp_b", self.species, "J/(mol*K**2)")

        d_h = (cp_a + 0.5 * d_T * cp_b) * d_T
        d_s = (cp_a - cp_b * T_ref) * log(f_T) + cp_b * d_T
        res["S"] += d_s.T @ n
        res["mu"] += d_h - T * d_s

        self.add_bound("T", T)  # logarithm taken


class StandardState(ThermoContribution):
    """This contribution assumes the current state to be the standard state
    and tags the following properties:

    ========== =========
    Property   Origin
    ========== =========
    ``S_std``  ``S``
    ``p_std``  ``p_ref``
    ``mu_std`` ``mu``
    ========== =========

    The contribution does not support any options or parameters.
    """

    provides = ["S_std", "p_std", "mu_std"]

    def define(self, res):
        # tag current chemical potential and entropy as standard state
        # create copy, such that tagged objects will not be mutated.
        res["S_std"] = copy(res["S"])
        res["p_std"] = copy(res["p_ref"])
        res["mu_std"] = copy(res["mu"])

        self.declare_vector_keys("mu_std")  # self.species can be default


class IdealMix(ThermoContribution):
    r"""This contribution supplies the ideal mix entropy contribution,
    applicable for both liquid and gas phases. It does not require any
    parameters nor supports options.


    The contribution does not define new properties, but builds on entropy
    ``S`` (:math:`S`) and chemical potential ``mu`` (:math:`\mu`). With
    :math:`\Delta s_i := -R\,\ln n_i/N`, we update

    .. math::

        S &\leftarrow S + \sum_i \Delta s_i\,n_i\\
        \mu_i &\leftarrow \mu_i - T\,\Delta s_i

    The contribution model domain is limited to positive quantities in order to
    prevent negative arguments to the logarithmic functions. One could enable
    also the third quadrant (**all** mole numbers negative), but doing so and
    thereby allowing the solver to jump between these two nearly disconnected
    domains has proven to be challenging in terms of solver robustness.
    """

    def define(self, res):
        T, n = res["T"], res["n"]
        N = qsum(n)
        x = n / N
        gtn = R_GAS * log(x)
        res["S"] -= n.T @ gtn
        res["mu"] += T * gtn

        self.add_bound("n", n, self.species)


class GibbsIdealGas(ThermoContribution):
    r"""This contribution supplements the ideal gas entropy contribution and
    defines the volume property in Gibbs coordinates, i.e. with volume as a
    function of pressure. The main use of this class is to actually represent
    an ideal gas model, while most non-ideal gas phase models are equations
    of state and formulated in Helmholtz coordinates.

    Based on the previously provided reference pressure ``p_ref`` and
    :math:`\Delta s := -R\,\ln p/p_\mathrm{ref}`, the update is

    .. math::

        S &\leftarrow S + \Delta s\,N\\
        \mu_i &\leftarrow \mu_i - T\,\Delta s

    Volume is defined as

    .. math::

        V = \frac{N\,R\,T}{p}

    Due to the logarithmic term in entropy, the contribution model domain is
    limited to positive pressures.
    """

    provides = ["V"]

    def define(self, res):
        T, p, n, p_ref = res["T"], res["p"], res["n"], res["p_ref"]
        N = qsum(n)
        gtn = R_GAS * log(p / p_ref)

        res["S"] -= N * gtn
        res["V"] = N * R_GAS * T / p
        res["mu"] += T * gtn

        self.add_bound("p", p)


class HelmholtzIdealGas(ThermoContribution):
    r"""This contribution supplements the ideal gas entropy contribution and
    defines the pressure property in Helmholtz coordinates, i.e. with pressure
    as function of volume. This is the common base contribution for most
    equations of state. Based on the previously provided reference pressure
    ``p_ref`` and

    .. math::
        \Delta s := -R\,\ln \frac{N\,R\,T}{V\,p_\mathrm{ref}}

    the update is

    .. math::

        S &\leftarrow S + \Delta s\,N\\
        \mu_i &\leftarrow \mu_i - T\,\Delta s

    Pressure is defined as

    .. math::

        p = \frac{N\,R\,T}{V}

    Due to the logarithmic term in entropy, the contribution model domain is
    limited to positive volumes.
    """

    def define(self, res):
        T, V, n, p_ref = res["T"], res["V"], res["n"], res["p_ref"]
        N = qsum(n)
        p = N * R_GAS * T / V
        gtn = R_GAS * log(p / p_ref)

        res["S"] -= N * gtn
        res["p"] = p
        res["mu"] += T * gtn

        self.add_bound("V", V)

    def initial_state(self, state, properties):
        volume = qsum(state.mol_vector) * R_GAS * \
                 state.temperature / state.pressure
        return ([base_magnitude(state.temperature),
                 base_magnitude(volume)] +
                list(base_magnitude(state.mol_vector)))


class ConstantGibbsVolume(ThermoContribution):
    r"""This contribution defines a mixture with constant molar volumes and
    hence zero compressibility and thermal expansion. This is an easy and
    sufficient way to describe a condensed phase at moderate pressures in
    limited temperature ranges. Mixing volume is not considered either, but can
    be added via additional contributions.

    The sole parameter is the molar volume for each species:

    ========= =================== ===========
    Parameter Description         Symbol
    ========= =================== ===========
    ``v_n``   Molar volume vector :math:`v_n`
    ========= =================== ===========

    The system volume is then calculated as

    .. math::

        V = \sum_i v_{n,i}\,n_i

    For the chemical potential, this yields, given pressure ``p`` and reference
    pressure ``p_ref``:

    .. math::

        \mu_i \leftarrow \mu_i + v_{n,i}\,(p - p_\mathrm{ref})

    .. note::

        The volume definition has a slight impact on the chemical potential,
        as shown in above equation. For a typical condensed phase however, say
        water, :math:`v_n = 1.8\cdot 10^{-5}\ \mathrm{m^3/mol}`, and if
        pressures are in the range of few bars,
        :math:`\Delta \mu < 10\ \mathrm{J/mol}`. This yields a 0.4 % change in
        the activity coefficient.
    """
    provides = ["V"]

    def define(self, res):
        n, p, p_ref = res["n"], res["p"], res["p_ref"]
        v_n = self.par_vector("v_n", self.species, "m**3/mol")
        res["mu"] += v_n * (p - p_ref)
        res["V"] = v_n.T @ n


class MolecularWeight(ThermoContribution):
    """This contribution defines the molecular weights as a species vector,
    based on the underlying :class:`simu.SpeciesDefinition` definitions."""
    provides = ["mw"]

    def define(self, res):
        mw = qvertcat(*[s.molecular_weight
                        for s in self.species_definitions.values()])
        res["mw"] = mw
        self.declare_vector_keys("mw")


class ChargeBalance(ThermoContribution):
    r"""This contribution defines a charge balance residual, if there are
    charged species. A ValueError is raised if only either positive or negative
    charged species exist.

    Given the presents of opposite charged species, the residual defined, based
    on the molar quantities :math:`n_i` is

    .. math:: \sum_i n_i \cdot c_i = 0

    Here, the charges :math:`c_i` are extracted from the species
    definition.

    By default, the unit for the residual is ``e`` for states and ``e / s`` for
    flows. This can be altered by the following options:

      - The option ``state_unit`` defines the unit used for states. If
        ``flow_unit`` is not given, it will be defined as ``state_unit / s``.
      - The option ``flow_unit`` defines the unit used for flows. It does not
        impact the ``state_unit`` value.

    """

    def define(self, res):
        charges = [s.charge for s in self.species_definitions.values()]
        pos = len([c for c in charges if c > 0])
        neg = len([c for c in charges if c < 0])
        if pos == 0 and neg == 0:
            return
        if pos == 0 or neg == 0:
            raise ValueError("Charges of only one sign cannot be balanced")
        charges = qvertcat(*charges)
        residual = res["n"].T  @ charges

        state_unit = self.options.get("state_unit", "e")
        flow_unit = self.options.get("flow_unit", f"{state_unit} / s")
        try:
            self.add_residual("balance", residual, flow_unit)
        except DimensionalityError:
            self.add_residual("balance", residual, state_unit)
